import asyncio
import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from decimal import Decimal
from typing import Any, Awaitable, Callable, Mapping, Optional, cast
from uuid import UUID, uuid4

import aiotools
import msgpack
import sqlalchemy as sa
from dateutil.tz import tzutc
from redis.asyncio import Redis
from redis.asyncio.client import Pipeline as RedisPipeline
from sqlalchemy.engine import Result, Row
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import joinedload, load_only, noload
from sqlalchemy.sql.expression import bindparam

from ai.backend.common import redis_helper
from ai.backend.common.types import RedisConnectionInfo, VFolderID
from ai.backend.common.utils import nmget
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.api.exceptions import VFolderOperationFailed
from ai.backend.manager.defs import DEFAULT_KEYPAIR_RATE_LIMIT, DEFAULT_KEYPAIR_RESOURCE_POLICY_NAME
from ai.backend.manager.models import (
    AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES,
    VFolderDeletionInfo,
    VFolderRow,
    VFolderStatusSet,
    initiate_vfolder_deletion,
    kernels,
    keypairs,
    vfolder_invitations,
    vfolder_permissions,
    vfolder_status_map,
    vfolders,
)
from ai.backend.manager.models.endpoint import (
    EndpointLifecycle,
    EndpointRow,
    EndpointTokenRow,
)
from ai.backend.manager.models.error_logs import error_logs
from ai.backend.manager.models.gql_models.keypair import CreateKeyPair
from ai.backend.manager.models.group import ProjectType, association_groups_users, groups
from ai.backend.manager.models.kernel import RESOURCE_USAGE_KERNEL_STATUSES
from ai.backend.manager.models.keypair import KeyPairRow
from ai.backend.manager.models.session import (
    AGENT_RESOURCE_OCCUPYING_SESSION_STATUSES,
    SessionRow,
)
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.user import UserRole, UserRow, UserStatus, users
from ai.backend.manager.models.utils import (
    ExtendedAsyncSAEngine,
    SAConnection,
    execute_with_retry,
    execute_with_txn_retry,
)
from ai.backend.manager.services.user.actions.admin_month_stats import (
    AdminMonthStatsAction,
    AdminMonthStatsActionResult,
)
from ai.backend.manager.services.user.actions.create_user import (
    CreateUserAction,
    CreateUserActionResult,
)
from ai.backend.manager.services.user.actions.delete_user import (
    DeleteUserAction,
    DeleteUserActionResult,
)
from ai.backend.manager.services.user.actions.modify_user import (
    ModifyUserAction,
    ModifyUserActionResult,
)
from ai.backend.manager.services.user.actions.purge_user import (
    PurgeUserAction,
    PurgeUserActionResult,
)
from ai.backend.manager.services.user.actions.user_month_stats import (
    UserMonthStatsAction,
    UserMonthStatsActionResult,
)
from ai.backend.manager.services.user.type import UserData

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class MutationResult:
    success: bool
    message: str
    data: Optional[Any]


class UserService:
    _db: ExtendedAsyncSAEngine
    _storage_manager: StorageSessionManager
    _redis_stat: RedisConnectionInfo

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        storage_manager: StorageSessionManager,
        redis_stat: RedisConnectionInfo,
    ) -> None:
        self._db = db
        self._storage_manager = storage_manager
        self._redis_stat = redis_stat

    async def create_user(self, action: CreateUserAction) -> CreateUserActionResult:
        username = action.input.username if action.input.username else action.input.email
        _status = UserStatus.ACTIVE  # TODO: Need to be set in action explicitly not in service (integrate is_active and status)
        if action.input.status is None and action.input.is_active is not None:
            _status = UserStatus.ACTIVE if action.input.is_active else UserStatus.INACTIVE
        if action.input.status is not None:
            _status = action.input.status
        group_ids = [] if action.input.group_ids is None else action.input.group_ids

        user_data = {
            "username": username,
            "email": action.input.email,
            "password": action.input.password,
            "need_password_change": action.input.need_password_change,
            "full_name": action.input.full_name,
            "description": action.input.description,
            "status": _status,
            "status_info": "admin-requested",  # user mutation is only for admin
            "domain_name": action.input.domain_name,
            "role": action.input.role,
            "allowed_client_ip": action.input.allowed_client_ip,
            "totp_activated": action.input.totp_activated,
            "resource_policy": action.input.resource_policy,
            "sudo_session_enabled": action.input.sudo_session_enabled,
        }
        if action.input.container_uid is not None:
            user_data["container_uid"] = action.input.container_uid
        if action.input.container_main_gid is not None:
            user_data["container_main_gid"] = action.input.container_main_gid
        if action.input.container_gids is not None:
            user_data["container_gids"] = action.input.container_gids

        user_insert_query = sa.insert(users).values(user_data)

        async def _post_func(conn: SAConnection, result: Result) -> Optional[Row]:
            if result.rowcount == 0:
                return None
            created_user = result.first()

            # Create a default keypair for the user.
            email = action.input.email
            kp_data = CreateKeyPair.prepare_new_keypair(
                email,
                {
                    "is_active": _status == UserStatus.ACTIVE,
                    "is_admin": user_data["role"] in [UserRole.SUPERADMIN, UserRole.ADMIN],
                    "resource_policy": DEFAULT_KEYPAIR_RESOURCE_POLICY_NAME,
                    "rate_limit": DEFAULT_KEYPAIR_RATE_LIMIT,
                },
            )
            kp_insert_query = sa.insert(keypairs).values(
                **kp_data,
                user=created_user.uuid,
            )
            await conn.execute(kp_insert_query)

            # Update user main_keypair
            main_ak = kp_data["access_key"]
            update_query = (
                sa.update(users)
                .where(users.c.uuid == created_user.uuid)
                .values(main_access_key=main_ak)
            )
            await conn.execute(update_query)

            model_store_query = sa.select([groups.c.id]).where(
                groups.c.type == ProjectType.MODEL_STORE
            )
            model_store_project = cast(
                dict[str, Any] | None, (await conn.execute(model_store_query)).first()
            )
            if model_store_project is not None:
                gids_to_join = [*group_ids, model_store_project["id"]]
            else:
                gids_to_join = group_ids

            # Add user to groups if group_ids parameter is provided.
            if gids_to_join:
                query = (
                    sa.select([groups.c.id])
                    .select_from(groups)
                    .where(groups.c.domain_name == action.input.domain_name)
                    .where(groups.c.id.in_(gids_to_join))
                )
                grps = (await conn.execute(query)).all()
                if grps:
                    group_data = [
                        {"user_id": created_user.uuid, "group_id": grp.id} for grp in grps
                    ]
                    group_insert_query = sa.insert(association_groups_users).values(group_data)
                    await conn.execute(group_insert_query)

            return created_user

        async def _do_mutate() -> MutationResult:
            async with self._db.begin() as conn:
                query = user_insert_query.returning(user_insert_query.table)
                result = await conn.execute(query)
                created_user = await _post_func(conn, result)
                return MutationResult(
                    success=True if created_user is not None else False,
                    message="User created successfully",
                    data=created_user,
                )

        result: MutationResult = await self._db_mutation_wrapper(_do_mutate)
        return CreateUserActionResult(
            data=UserData.from_row(result.data),
            success=result.success,
        )

    async def modify_user(self, action: ModifyUserAction) -> ModifyUserActionResult:
        email = action.email
        data = action.modifier.fields_to_update()
        if data.get("password") is None:
            data.pop("password", None)

        group_ids = action.group_ids.optional_value()

        if not data and group_ids is None:
            return ModifyUserActionResult(data=None, success=False)
        if data.get("status") is None and data.get("is_active") is not None:
            data["status"] = UserStatus.ACTIVE if data["is_active"] else UserStatus.INACTIVE

        if data.get("password") is not None:
            data["password_changed_at"] = sa.func.now()

        main_access_key: str | None = data.get("main_access_key")
        user_update_data: dict[str, Any] = {}
        prev_domain_name: str
        prev_role: UserRole

        async def _pre_func(conn: SAConnection) -> None:
            nonlocal user_update_data, prev_domain_name, prev_role, main_access_key
            result = await conn.execute(
                sa.select([users.c.domain_name, users.c.role, users.c.status])
                .select_from(users)
                .where(users.c.email == email),
            )
            row = result.first()
            prev_domain_name = row.domain_name
            prev_role = row.role
            user_update_data = data.copy()
            if "status" in data and row.status != data["status"]:
                user_update_data["status_info"] = (
                    "admin-requested"  # user mutation is only for admin
                )
            if main_access_key is not None:
                db_session = SASession(conn)
                keypair_query = (
                    sa.select(KeyPairRow)
                    .where(KeyPairRow.access_key == main_access_key)
                    .options(
                        noload("*"),
                        joinedload(KeyPairRow.user_row).options(load_only(UserRow.email)),
                    )
                )
                keypair_row: KeyPairRow | None = (await db_session.scalars(keypair_query)).first()
                if keypair_row is None:
                    raise RuntimeError("Cannot set non-existing access key as the main access key.")
                if keypair_row.user_row.email != email:
                    raise RuntimeError(
                        "Cannot set another user's access key as the main access key."
                    )
                await conn.execute(
                    sa.update(users)
                    .where(users.c.email == email)
                    .values(main_access_key=main_access_key)
                )

        update_query = lambda: (  # uses lambda because user_update_data is modified in _pre_func()
            sa.update(users).values(user_update_data).where(users.c.email == email)
        )

        async def _post_func(conn: SAConnection, result: Result) -> Row:
            nonlocal prev_domain_name, prev_role
            updated_user = result.first()
            if "role" in data and data["role"] != prev_role:
                from ai.backend.manager.models import keypairs

                result = await conn.execute(
                    sa.select([
                        keypairs.c.user,
                        keypairs.c.is_active,
                        keypairs.c.is_admin,
                        keypairs.c.access_key,
                    ])
                    .select_from(keypairs)
                    .where(keypairs.c.user == updated_user.uuid)
                    .order_by(sa.desc(keypairs.c.is_admin))
                    .order_by(sa.desc(keypairs.c.is_active)),
                )
                if data["role"] in [UserRole.SUPERADMIN, UserRole.ADMIN]:
                    # User's becomes admin. Set the keypair as active admin.
                    # TODO: Should we update the role of all users related to keypair?
                    kp = result.first()
                    kp_data = dict()
                    if not kp.is_admin:
                        kp_data["is_admin"] = True
                    if not kp.is_active:
                        kp_data["is_active"] = True
                    if kp_data:
                        await conn.execute(
                            sa.update(keypairs)
                            .values(kp_data)
                            .where(keypairs.c.user == updated_user.uuid),
                        )
                else:
                    # User becomes non-admin. Make the keypair non-admin as well.
                    # If there are multiple admin keypairs, inactivate them.
                    # TODO: Should elaborate keypair inactivation policy.
                    rows = result.fetchall()
                    kp_updates = []
                    for idx, row in enumerate(rows):
                        kp_data = {
                            "b_access_key": row.access_key,
                            "is_admin": row.is_admin,
                            "is_active": row.is_active,
                        }
                        if idx == 0:
                            kp_data["is_admin"] = False
                            kp_updates.append(kp_data)
                            continue
                        if row.is_admin and row.is_active:
                            kp_data["is_active"] = False
                            kp_updates.append(kp_data)
                    if kp_updates:
                        await conn.execute(
                            sa.update(keypairs)
                            .values({
                                "is_admin": bindparam("is_admin"),
                                "is_active": bindparam("is_active"),
                            })
                            .where(keypairs.c.access_key == bindparam("b_access_key")),
                            kp_updates,
                        )

            # If domain is changed and no group is associated, clear previous domain's group.
            if prev_domain_name != updated_user.domain_name and not group_ids:
                await conn.execute(
                    sa.delete(association_groups_users).where(
                        association_groups_users.c.user_id == updated_user.uuid
                    ),
                )

            # Update user's group if group_ids parameter is provided.
            if group_ids and updated_user is not None:
                # Clear previous groups associated with the user.
                await conn.execute(
                    sa.delete(association_groups_users).where(
                        association_groups_users.c.user_id == updated_user.uuid
                    ),
                )
                # Add user to new groups.
                result = await conn.execute(
                    sa.select([groups.c.id])
                    .select_from(groups)
                    .where(groups.c.domain_name == updated_user.domain_name)
                    .where(groups.c.id.in_(group_ids)),
                )
                grps = result.fetchall()
                if grps:
                    values = [{"user_id": updated_user.uuid, "group_id": grp.id} for grp in grps]
                    await conn.execute(
                        sa.insert(association_groups_users).values(values),
                    )

            return updated_user

        async def _do_mutate() -> MutationResult:
            async with self._db.begin() as conn:
                await _pre_func(conn)
                query = update_query().returning(update_query().table)
                result = await conn.execute(query)
                updated_user = await _post_func(conn, result)

            return MutationResult(
                success=True,
                message="User modified successfully",
                data=updated_user,
            )

        result: MutationResult = await self._db_mutation_wrapper(_do_mutate)

        return ModifyUserActionResult(
            success=result.success,
            data=UserData.from_row(result.data),
        )

    async def delete_user(self, action: DeleteUserAction) -> DeleteUserActionResult:
        async def _pre_func(conn: SAConnection) -> None:
            # Make all user keypairs inactive.
            await conn.execute(
                sa.update(keypairs)
                .values(is_active=False)
                .where(keypairs.c.user_id == action.email),
            )

        update_query = (
            sa.update(users)
            .values(status=UserStatus.DELETED, status_info="admin-requested")
            .where(users.c.email == action.email)
        )

        async def _do_mutate() -> MutationResult:
            async with self._db.begin() as conn:
                await _pre_func(conn)
                await conn.execute(update_query)

            return MutationResult(
                success=True,
                message="User deleted successfully",
                data=None,
            )

        result: MutationResult = await self._db_mutation_wrapper(_do_mutate)
        return DeleteUserActionResult(
            success=result.success,
        )

    async def purge_user(self, action: PurgeUserAction) -> PurgeUserActionResult:
        email = action.email

        async def _delete(db_session: SASession) -> None:
            conn = await db_session.connection()
            user_uuid = await db_session.scalar(
                sa.select(UserRow.uuid).where(UserRow.email == email),
            )
            user_uuid = cast(Optional[UUID], user_uuid)
            log.info("Purging all records of the user {0}...", email)
            if user_uuid is None:
                raise RuntimeError(f"User not found (email: {email})")

            if await self.user_vfolder_mounted_to_active_kernels(conn, user_uuid):
                raise RuntimeError(
                    "Some of user's virtual folders are mounted to active kernels. "
                    "Terminate those kernels first.",
                )

            if action.purge_shared_vfolders.optional_value():
                await self.migrate_shared_vfolders(
                    conn,
                    deleted_user_uuid=user_uuid,
                    target_user_uuid=action.user_info_ctx.uuid,
                    target_user_email=action.user_info_ctx.email,
                )
            if action.delegate_endpoint_ownership.optional_value():
                await EndpointRow.delegate_endpoint_ownership(
                    db_session,
                    user_uuid,
                    action.user_info_ctx.uuid,
                    action.user_info_ctx.main_access_key,
                )
                await self._delete_endpoint(db_session, user_uuid, delete_destroyed_only=True)
            else:
                await self._delete_endpoint(db_session, user_uuid, delete_destroyed_only=False)
            if await self._user_has_active_sessions(db_session, user_uuid):
                raise RuntimeError("User has some active sessions. Terminate them first.")
            await self._delete_sessions(db_session, user_uuid)
            await self._delete_vfolders(self._db, user_uuid, self._storage_manager)
            await self.delete_error_logs(conn, user_uuid)
            await self.delete_keypairs(conn, self._redis_stat, user_uuid)

            await db_session.execute(sa.delete(users).where(users.c.email == email))

        async with self._db.connect() as db_conn:
            await execute_with_txn_retry(_delete, self._db.begin_session, db_conn)

        return PurgeUserActionResult(success=True)

    async def migrate_shared_vfolders(
        self,
        conn: SAConnection,
        deleted_user_uuid: UUID,
        target_user_uuid: UUID,
        target_user_email: str,
    ) -> int:
        """
        Migrate shared virtual folders' ownership to a target user.

        If migrating virtual folder's name collides with target user's already
        existing folder, append random string to the migrating one.

        :param conn: DB connection
        :param deleted_user_uuid: user's UUID who will be deleted
        :param target_user_uuid: user's UUID who will get the ownership of virtual folders

        :return: number of deleted rows
        """
        # Gather target user's virtual folders' names.
        query = (
            sa.select([vfolders.c.name])
            .select_from(vfolders)
            .where(vfolders.c.user == target_user_uuid)
        )
        existing_vfolder_names = [row.name async for row in (await conn.stream(query))]

        # Migrate shared virtual folders.
        # If virtual folder's name collides with target user's folder,
        # append random string to the name of the migrating folder.
        j = vfolder_permissions.join(
            vfolders,
            vfolder_permissions.c.vfolder == vfolders.c.id,
        )
        query = (
            sa.select([vfolders.c.id, vfolders.c.name])
            .select_from(j)
            .where(vfolders.c.user == deleted_user_uuid)
        )
        migrate_updates = []
        async for row in await conn.stream(query):
            name = row.name
            if name in existing_vfolder_names:
                name += f"-{uuid4().hex[:10]}"
            migrate_updates.append({"vid": row.id, "vname": name})
        if migrate_updates:
            # Remove invitations and vfolder_permissions from target user.
            # Target user will be the new owner, and it does not make sense to have
            # invitation and shared permission for its own folder.
            migrate_vfolder_ids = [item["vid"] for item in migrate_updates]
            delete_query = sa.delete(vfolder_invitations).where(
                (vfolder_invitations.c.invitee == target_user_email)
                & (vfolder_invitations.c.vfolder.in_(migrate_vfolder_ids))
            )
            await conn.execute(delete_query)
            delete_query = sa.delete(vfolder_permissions).where(
                (vfolder_permissions.c.user == target_user_uuid)
                & (vfolder_permissions.c.vfolder.in_(migrate_vfolder_ids))
            )
            await conn.execute(delete_query)

            rowcount = 0
            for item in migrate_updates:
                update_query = (
                    sa.update(vfolders)
                    .values(
                        user=target_user_uuid,
                        name=item["vname"],
                    )
                    .where(vfolders.c.id == item["vid"])
                )
                result = await conn.execute(update_query)
                rowcount += result.rowcount
            if rowcount > 0:
                log.info(
                    "{0} shared folders are detected and migrated to user {1}",
                    rowcount,
                    target_user_uuid,
                )
            return rowcount
        else:
            return 0

    async def _delete_vfolders(
        self,
        engine: ExtendedAsyncSAEngine,
        user_uuid: UUID,
        storage_manager: StorageSessionManager,
        *,
        delete_service: bool = False,
    ) -> int:
        """
        Delete user's all virtual folders as well as their physical data.

        :param conn: DB connection
        :param user_uuid: user's UUID to delete virtual folders

        :return: number of deleted rows
        """
        target_vfs: list[VFolderDeletionInfo] = []
        async with engine.begin_session() as db_session:
            await db_session.execute(
                vfolder_permissions.delete().where(vfolder_permissions.c.user == user_uuid),
            )
            result = await db_session.scalars(
                sa.select(VFolderRow).where(
                    sa.and_(
                        VFolderRow.user == user_uuid,
                        VFolderRow.status.in_(vfolder_status_map[VFolderStatusSet.DELETABLE]),
                    )
                ),
            )
            rows = cast(list[VFolderRow], result.fetchall())
            for vf in rows:
                target_vfs.append(
                    VFolderDeletionInfo(VFolderID.from_row(vf), vf.host, vf.unmanaged_path)
                )

        storage_ptask_group = aiotools.PersistentTaskGroup()
        try:
            await initiate_vfolder_deletion(
                engine,
                target_vfs,
                storage_manager,
                storage_ptask_group,
            )
        except VFolderOperationFailed as e:
            log.error("error on deleting vfolder filesystem directory: {0}", e.extra_msg)
            raise
        deleted_count = len(target_vfs)
        if deleted_count > 0:
            log.info("deleted {0} user's virtual folders ({1})", deleted_count, user_uuid)
        return deleted_count

    async def user_vfolder_mounted_to_active_kernels(
        self,
        conn: SAConnection,
        user_uuid: UUID,
    ) -> bool:
        """
        Check if no active kernel is using the user's virtual folders.

        :param conn: DB connection
        :param user_uuid: user's UUID

        :return: True if a virtual folder is mounted to active kernels.
        """

        result = await conn.execute(
            sa.select([vfolders.c.id]).select_from(vfolders).where(vfolders.c.user == user_uuid),
        )
        rows = result.fetchall()
        user_vfolder_ids = [row.id for row in rows]
        query = (
            sa.select([kernels.c.mounts])
            .select_from(kernels)
            .where(kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES))
        )
        async for row in await conn.stream(query):
            for _mount in row["mounts"]:
                try:
                    vfolder_id = UUID(_mount[2])
                    if vfolder_id in user_vfolder_ids:
                        return True
                except Exception:
                    pass
        return False

    async def _user_has_active_sessions(
        self,
        db_session: SASession,
        user_uuid: UUID,
    ) -> bool:
        """
        Check if the user does not have active sessions.
        """
        active_session_count = await db_session.scalar(
            sa.select(sa.func.count())
            .select_from(SessionRow)
            .where(
                sa.and_(
                    SessionRow.user_uuid == user_uuid,
                    SessionRow.status.in_(AGENT_RESOURCE_OCCUPYING_SESSION_STATUSES),
                )
            ),
        )
        return active_session_count > 0

    async def _delete_endpoint(
        self,
        db_session: SASession,
        user_uuid: UUID,
        *,
        delete_destroyed_only: bool = False,
    ) -> None:
        """
        Delete user's all endpoint.
        """
        if delete_destroyed_only:
            status_filter = {EndpointLifecycle.DESTROYED}
        else:
            status_filter = {status for status in EndpointLifecycle}
        endpoint_rows = await EndpointRow.list(
            db_session, user_uuid=user_uuid, load_tokens=True, status_filter=status_filter
        )
        token_ids_to_delete = []
        endpoint_ids_to_delete = []
        for row in endpoint_rows:
            token_ids_to_delete.extend([token.id for token in row.tokens])
            endpoint_ids_to_delete.append(row.id)
        await db_session.execute(
            sa.delete(EndpointTokenRow).where(EndpointTokenRow.id.in_(token_ids_to_delete))
        )
        await db_session.execute(
            sa.delete(EndpointRow).where(EndpointRow.id.in_(endpoint_ids_to_delete))
        )

    async def delete_error_logs(
        self,
        conn: SAConnection,
        user_uuid: UUID,
    ) -> int:
        """
        Delete user's all error logs.

        :param conn: DB connection
        :param user_uuid: user's UUID to delete error logs
        :return: number of deleted rows
        """
        result = await conn.execute(sa.delete(error_logs).where(error_logs.c.user == user_uuid))
        if result.rowcount > 0:
            log.info("deleted {0} user's error logs ({1})", result.rowcount, user_uuid)
        return result.rowcount

    async def _delete_sessions(
        self,
        db_session: SASession,
        user_uuid: UUID,
    ) -> None:
        """
        Delete user's sessions.
        """
        from ai.backend.manager.models.session import SessionRow

        await SessionRow.delete_by_user_id(user_uuid, db_session=db_session)

    async def delete_keypairs(
        self,
        conn: SAConnection,
        redis_conn: RedisConnectionInfo,
        user_uuid: UUID,
    ) -> int:
        """
        Delete user's all keypairs.

        :param conn: DB connection
        :param redis_conn: redis connection info
        :param user_uuid: user's UUID to delete keypairs
        :return: number of deleted rows
        """
        ak_rows = await conn.execute(
            sa.select([keypairs.c.access_key]).where(keypairs.c.user == user_uuid),
        )
        if (row := ak_rows.first()) and (access_key := row.access_key):
            # Log concurrency used only when there is at least one keypair.
            await redis_helper.execute(
                redis_conn,
                lambda r: r.delete(f"keypair.concurrency_used.{access_key}"),
            )
            await redis_helper.execute(
                redis_conn,
                lambda r: r.delete(f"keypair.sftp_concurrency_used.{access_key}"),
            )
        result = await conn.execute(
            sa.delete(keypairs).where(keypairs.c.user == user_uuid),
        )
        if result.rowcount > 0:
            log.info("deleted {0} user's keypairs ({1})", result.rowcount, user_uuid)
        return result.rowcount

    async def _db_mutation_wrapper(
        self, _do_mutate: Callable[[], Awaitable[MutationResult]]
    ) -> MutationResult:
        try:
            return await execute_with_retry(_do_mutate)
        except sa.exc.IntegrityError as e:
            log.warning("db_mutation_wrapper(): integrity error ({})", repr(e))
            return MutationResult(success=False, message=f"integrity error: {e}", data=None)
        except sa.exc.StatementError as e:
            log.warning(
                "db_mutation_wrapper(): statement error ({})\n{}",
                repr(e),
                e.statement or "(unknown)",
            )
            orig_exc = e.orig
            return MutationResult(success=False, message=str(orig_exc), data=None)
        except (asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except Exception:
            log.exception("db_mutation_wrapper(): other error")
            raise

    async def _get_time_binned_monthly_stats(self, user_uuid=None):
        """
        Generate time-binned (15 min) stats for the last one month (2880 points).
        The structure of the result would be:

            [
            # [
            #     timestamp, num_sessions,
            #     cpu_allocated, mem_allocated, gpu_allocated,
            #     io_read, io_write, scratch_used,
            # ]
                [1562083808.657106, 1, 1.2, 1073741824, ...],
                [1562084708.657106, 2, 4.0, 1073741824, ...],
            ]

        Note that the timestamp is in UNIX-timestamp.
        """
        # Get all or user kernels for the last month from DB.
        time_window = 900  # 15 min
        stat_length = 2880  # 15 * 4 * 24 * 30
        now = datetime.now(tzutc())
        start_date = now - timedelta(days=30)

        async with self._db.begin_readonly() as conn:
            query = (
                sa.select([
                    kernels.c.id,
                    kernels.c.created_at,
                    kernels.c.terminated_at,
                    kernels.c.occupied_slots,
                ])
                .select_from(kernels)
                .where(
                    (kernels.c.terminated_at >= start_date)
                    & (kernels.c.status.in_(RESOURCE_USAGE_KERNEL_STATUSES)),
                )
                .order_by(sa.asc(kernels.c.created_at))
            )
            if user_uuid is not None:
                query = query.where(kernels.c.user_uuid == user_uuid)
            result = await conn.execute(query)
            rows = result.fetchall()

        # Build time-series of time-binned stats.
        start_date_ts = start_date.timestamp()
        time_series_list: list[dict[str, Any]] = [
            {
                "date": start_date_ts + (idx * time_window),
                "num_sessions": {
                    "value": 0,
                    "unit_hint": "count",
                },
                "cpu_allocated": {
                    "value": 0,
                    "unit_hint": "count",
                },
                "mem_allocated": {
                    "value": 0,
                    "unit_hint": "bytes",
                },
                "gpu_allocated": {
                    "value": 0,
                    "unit_hint": "count",
                },
                "io_read_bytes": {
                    "value": 0,
                    "unit_hint": "bytes",
                },
                "io_write_bytes": {
                    "value": 0,
                    "unit_hint": "bytes",
                },
                "disk_used": {
                    "value": 0,
                    "unit_hint": "bytes",
                },
            }
            for idx in range(stat_length)
        ]

        async def _pipe_builder(r: Redis) -> RedisPipeline:
            pipe = r.pipeline()
            for row in rows:
                await pipe.get(str(row["id"]))
            return pipe

        raw_stats = await redis_helper.execute(self._redis_stat, _pipe_builder)

        for row, raw_stat in zip(rows, raw_stats):
            if raw_stat is not None:
                last_stat = msgpack.unpackb(raw_stat)
                io_read_byte = int(nmget(last_stat, "io_read.current", 0))
                io_write_byte = int(nmget(last_stat, "io_write.current", 0))
                disk_used = int(nmget(last_stat, "io_scratch_size.stats.max", 0, "/"))
            else:
                io_read_byte = 0
                io_write_byte = 0
                disk_used = 0

            occupied_slots: Mapping[str, Any] = row.occupied_slots
            kernel_created_at: float = row.created_at.timestamp()
            kernel_terminated_at: float = row.terminated_at.timestamp()
            cpu_value = int(occupied_slots.get("cpu", 0))
            mem_value = int(occupied_slots.get("mem", 0))
            cuda_device_value = int(occupied_slots.get("cuda.devices", 0))
            cuda_share_value = Decimal(occupied_slots.get("cuda.shares", 0))

            start_index = int((kernel_created_at - start_date_ts) // time_window)
            end_index = int((kernel_terminated_at - start_date_ts) // time_window) + 1
            if start_index < 0:
                start_index = 0
            for time_series in time_series_list[start_index:end_index]:
                time_series["num_sessions"]["value"] += 1
                time_series["cpu_allocated"]["value"] += cpu_value
                time_series["mem_allocated"]["value"] += mem_value
                time_series["gpu_allocated"]["value"] += cuda_device_value
                time_series["gpu_allocated"]["value"] += cuda_share_value
                time_series["io_read_bytes"]["value"] += io_read_byte
                time_series["io_write_bytes"]["value"] += io_write_byte
                time_series["disk_used"]["value"] += disk_used

        # Change Decimal type to float to serialize to JSON
        for time_series in time_series_list:
            time_series["gpu_allocated"]["value"] = float(time_series["gpu_allocated"]["value"])
        return time_series_list

    async def user_month_stats(self, action: UserMonthStatsAction) -> UserMonthStatsActionResult:
        stats = await self._get_time_binned_monthly_stats(user_uuid=action.user_id)
        return UserMonthStatsActionResult(stats=stats)

    # TODO: user (전체)
    async def admin_month_stats(self, action: AdminMonthStatsAction) -> AdminMonthStatsActionResult:
        stats = await self._get_time_binned_monthly_stats(user_uuid=None)
        return AdminMonthStatsActionResult(stats=stats)



from abc import ABC

from ..action import BaseAction, BaseActionTriggerMeta, ProcessResult


class ActionMonitor(ABC):
    async def prepare(self, action: BaseAction, meta: BaseActionTriggerMeta) -> None:
        pass

    async def done(self, action: BaseAction, result: ProcessResult) -> None:
        pass



from __future__ import annotations

import traceback
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Literal,
    NotRequired,
    Optional,
    Tuple,
    TypedDict,
)

from aiotools import TaskGroupError

if TYPE_CHECKING:
    from ai.backend.common.types import AgentId


class InvalidArgument(Exception):
    """
    An internal exception class to represent invalid arguments in internal APIs.
    This is wrapped as InvalidAPIParameters in web request handlers.
    """

    pass


class RPCError(RuntimeError):
    """
    An exception class to represent any error caused in RPC functions.
    """

    __slots__ = (
        "agent_id",
        "agent_addr",
        "extra_msg",
    )

    def __init__(
        self,
        agent_id: AgentId,
        agent_addr: str,
        extra_msg: str,
    ) -> None:
        super().__init__(agent_id, agent_addr, extra_msg)
        self.agent_id = agent_id
        self.agent_addr = agent_addr
        self.extra_msg = extra_msg


class AgentError(RuntimeError):
    """
    A dummy exception class to distinguish agent-side errors passed via
    agent rpc calls.

    It carries two args tuple: the exception type and exception arguments from
    the agent.
    """

    __slots__ = (
        "agent_id",
        "exc_name",
        "exc_repr",
        "exc_tb",
    )

    def __init__(
        self,
        agent_id: AgentId,
        exc_name: str,
        exc_repr: str,
        exc_args: Tuple[Any, ...],
        exc_tb: Optional[str] = None,
    ) -> None:
        super().__init__(agent_id, exc_name, exc_repr, exc_args, exc_tb)
        self.agent_id = agent_id
        self.exc_name = exc_name
        self.exc_repr = exc_repr
        self.exc_args = exc_args
        self.exc_tb = exc_tb


class MultiAgentError(TaskGroupError):
    """
    An exception that is a collection of multiple errors from multiple agents.
    """


class ErrorDetail(TypedDict):
    src: str
    name: str
    repr: str
    agent_id: NotRequired[str]
    collection: NotRequired[List[ErrorDetail]]
    traceback: NotRequired[str]


class ErrorStatusInfo(TypedDict):
    error: ErrorDetail


def convert_to_status_data(
    e: BaseException,
    is_debug: bool = False,
    *,
    src: str | None = None,
) -> ErrorStatusInfo:
    data: ErrorStatusInfo
    match e:
        case MultiAgentError():
            return {
                "error": {
                    "src": "agent",
                    "name": "MultiAgentError",
                    "repr": f"MultiAgentError({len(e.__errors__)})",
                    "collection": [
                        convert_to_status_data(sub_error, is_debug, src="agent")["error"]
                        for sub_error in e.__errors__
                    ],
                },
            }
        case AgentError():
            data = {
                "error": {
                    "src": "agent",
                    "name": e.exc_name,
                    "repr": e.exc_repr,
                },
            }
            if is_debug:
                data["error"]["agent_id"] = e.agent_id
                data["error"]["traceback"] = e.exc_tb or ""
            return data
        case _:
            data = {
                "error": {
                    "src": "other" if src is None else src,
                    "name": e.__class__.__name__,
                    "repr": repr(e),
                },
            }
            if is_debug:
                data["error"]["traceback"] = "\n".join(traceback.format_tb(e.__traceback__))
            return data


class ContainerRegistryProjectEmpty(RuntimeError):
    def __init__(self, type: str, project: Literal[""] | None):
        super().__init__(
            f"{type} container registry requires project value, but {project} is provided"
        )


class ScanImageError(RuntimeError):
    pass


class ScanTagError(RuntimeError):
    pass




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/manager:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.13',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    ],
    'description': 'Backend.AI Manager',
    'entry_points': {
        'backendai_cli_v10': [
            'mgr = ai.backend.manager.cli.__main__:main',
            'mgr.start-server = ai.backend.manager.server:main',
        ],
        'backendai_scheduler_v10': [
            'fifo = ai.backend.manager.scheduler.fifo:FIFOSlotScheduler',
            'lifo = ai.backend.manager.scheduler.fifo:LIFOSlotScheduler',
            'drf = ai.backend.manager.scheduler.drf:DRFScheduler',
        ],
        'backendai_agentselector_v10': [
            'legacy = ai.backend.manager.scheduler.agent_selector:LegacyAgentSelector',
            'roundrobin = ai.backend.manager.scheduler.agent_selector:RoundRobinAgentSelector',
            'concentrated = ai.backend.manager.scheduler.agent_selector:ConcentratedAgentSelector',
            'dispersed = ai.backend.manager.scheduler.agent_selector:DispersedAgentSelector',
        ],
        'backendai_error_monitor_v20': [
            'intrinsic = ai.backend.manager.plugin.error_monitor:ErrorMonitor',
        ],
        'backendai_network_manager_v1': [
            'overlay = ai.backend.manager.network.overlay:OverlayNetworkPlugin',
        ],
    },
    'install_requires': (
        'PyJWT~=2.10.1',
        'PyYAML~=6.0',
        'SQLAlchemy[postgresql_asyncpg]~=1.4.54',
        'aiodataloader~=0.4.2',
        'aiodocker==0.24.0',
        'aiofiles~=24.1.0',
        'aiohttp_cors~=0.8.1',
        'aiohttp_sse>=2.2',
        'aiohttp~=3.11.16',
        'aiomonitor~=0.7.0',
        'aiotools~=1.9.0',
        'alembic~=1.13.2',
        'async_timeout~=4.0',
        'attrs>=25.3',
        'backend.ai-cli==25.6.15',
        'backend.ai-common==25.6.15',
        'backend.ai-logging==25.6.15',
        'backend.ai-plugin==25.6.15',
        'bcrypt~=4.2.0',
        'boto3~=1.35',
        'callosum~=1.0.3',
        'click~=8.1.7',
        'cryptography>=44.0.2',
        'etcd-client-py~=0.4.0',
        'graphene~=3.3.0',
        'lark~=1.1.5',
        'more-itertools~=10.5.0',
        'msgpack~=1.1.0',
        'multidict~=6.2.0',
        'pydantic[email]~=2.11.3',
        'python-dateutil>=2.9',
        'pyzmq~=26.4',
        'redis[hiredis]==4.5.5',
        'ruamel.yaml~=0.18.10',
        'setproctitle~=1.3.5',
        'tabulate~=0.8.9',
        'tenacity>=9.0',
        'trafaret~=2.1',
        'typeguard~=4.3',
        'types-PyYAML',
        'types-aiofiles',
        'types-python-dateutil',
        'types-redis',
        'types-six',
        'types-tabulate',
        'typing_extensions~=4.11',
        'uvloop~=0.21; sys_platform != "Windows"',
        'yarl~=1.19.0',
    ),
    'license': 'LGPLv3',
    'long_description': """Backend.AI Manager with API Gateway
===================================

Package Structure
-----------------

* `ai.backend.manager`: Computing resource and workload management with public APIs

Installation
------------

Please visit [the installation guides](https://github.com/lablup/backend.ai/wiki).


### Kernel/system configuration

#### Recommended resource limits:

**`/etc/security/limits.conf`**
```
root hard nofile 512000
root soft nofile 512000
root hard nproc 65536
root soft nproc 65536
user hard nofile 512000
user soft nofile 512000
user hard nproc 65536
user soft nproc 65536
```

**sysctl**
```
fs.file-max=2048000
net.core.somaxconn=1024
net.ipv4.tcp_max_syn_backlog=1024
net.ipv4.tcp_slow_start_after_idle=0
net.ipv4.tcp_fin_timeout=10
net.ipv4.tcp_window_scaling=1
net.ipv4.tcp_tw_reuse=1
net.ipv4.tcp_early_retrans=1
net.ipv4.ip_local_port_range="10000 65000"
net.core.rmem_max=16777216
net.core.wmem_max=16777216
net.ipv4.tcp_rmem=4096 12582912 16777216
net.ipv4.tcp_wmem=4096 12582912 16777216
```


### For development

#### Prerequisites

* Python 3.6 or higher with [pyenv](https://github.com/pyenv/pyenv)
and [pyenv-virtualenv](https://github.com/pyenv/pyenv-virtualenv) (optional but recommneded)
* Docker 18.03 or later with docker-compose (18.09 or later is recommended)

#### Common steps

Clone [the meta repository](https://github.com/lablup/backend.ai) and install a "halfstack"
configuration.  The halfstack configuration installs and runs several dependency daemons such as etcd in
the background.

```console
$ git clone https://github.com/lablup/backend.ai halfstack
$ cd halfstack
$ docker-compose -f docker-compose.halfstack.yml up -d
```

Then prepare the source clone of the agent as follows.
First install the current working copy.

```console
$ git clone https://github.com/lablup/backend.ai-manager manager
$ cd manager
$ pyenv virtualenv venv-manager
$ pyenv local venv-manager
$ pip install -U pip setuptools
$ pip install -U -r requirements/dev.txt
```

From now on, let's assume all shell commands are executed inside the virtualenv.

### Halfstack (single-node development & testing)

#### Recommended directory structure

* `backend.ai-dev`
  - `manager` (git clone from this repo)
  - `agent` (git clone from [the agent repo](https://github.com/lablup/backend.ai-agent))
  - `common` (git clone from [the common repo](https://github.com/lablup/backend.ai-common))

Install `backend.ai-common` as an editable package in the manager (and the agent) virtualenvs
to keep the codebase up-to-date.

```console
$ cd manager
$ pip install -U -e ../common -r requirements/dev.txt
```

#### Steps

Copy (or symlink) the halfstack configs:
```console
$ cp config/halfstack.toml ./manager.toml
$ cp config/halfstack.alembic.ini ./alembic.ini
```

Set up Redis:
```console
$ backend.ai mgr etcd put config/redis/addr 127.0.0.1:8110
```

> ℹ️ NOTE: You may replace `backend.ai mgr` with `python -m ai.backend.manager.cli` in case your `PATH` is unmodifiable.

Set up the public Docker registry:
```console
$ backend.ai mgr etcd put config/docker/registry/index.docker.io "https://registry-1.docker.io"
$ backend.ai mgr etcd put config/docker/registry/index.docker.io/username "lablup"
$ backend.ai mgr image rescan index.docker.io
```

Set up the vfolder paths:
```console
$ mkdir -p "$HOME/vfroot/local"
$ backend.ai mgr etcd put volumes/_mount "$HOME/vfroot"
$ backend.ai mgr etcd put volumes/_default_host local
```

Set up the allowed types of vfolder. Allowed values are "user" or "group".
If none is specified, "user" type is set implicitly:
```console
$ backend.ai mgr etcd put volumes/_types/user ""   # enable user vfolder
$ backend.ai mgr etcd put volumes/_types/group ""  # enable group vfolder
```

Set up the database:
```console
$ backend.ai mgr schema oneshot
$ backend.ai mgr fixture populate sample-configs/example-keypairs.json
$ backend.ai mgr fixture populate sample-configs/example-resource-presets.json
```

Then, run it (for debugging, append a `--debug` flag):

```console
$ backend.ai mgr start-server
```

To run tests:

```console
$ python -m flake8 src tests
$ python -m pytest -m 'not integration' tests
```

Now you are ready to install the agent.
Head to [the README of Backend.AI Agent](https://github.com/lablup/backend.ai-agent/blob/master/README.md).

NOTE: To run tests including integration tests, you first need to install and run the agent on the same host.

## Deployment

### Configuration

Put a TOML-formatted manager configuration (see the sample in `config/sample.toml`)
in one of the following locations:

 * `manager.toml` (current working directory)
 * `~/.config/backend.ai/manager.toml` (user-config directory)
 * `/etc/backend.ai/manager.toml` (system-config directory)

Only the first found one is used by the daemon.

Also many configurations shared by both manager and agent are stored in etcd.
As you might have noticed above, the manager provides a CLI interface to access and manipulate the etcd
data.  Check out the help page of our etcd command set:

```console
$ python -m ai.backend.manager.cli etcd --help
```

If you run etcd as a Docker container (e.g., via halfstack), you may use the native client as well.
In this case, PLEASE BE WARNED that you must prefix the keys with "/sorna/{namespace}" manaully:

```console
$ docker exec -it ${ETCD_CONTAINER_ID} /bin/ash -c 'ETCDCTL_API=3 etcdctl ...'
```

### Running from a command line

The minimal command to execute:

```sh
python -m ai.backend.gateway.server
```

For more arguments and options, run the command with `--help` option.

### Writing a wrapper script

To use with systemd, crontab, and other system-level daemons, you may need to write a shell script
that executes specific CLI commands provided by Backend.AI modules.

The following example shows how to set up pyenv and virtualenv for the script-local environment.
It runs the gateway server if no arguments are given, and execute the given arguments as a shell command
if any.
For instance, you may get/set configurations like: `run-manager.sh python -m ai.backend.manager.etcd ...`
where the name of scripts is `run-manager.sh`.

```bash
#! /bin/bash
if [ -z "$HOME" ]; then
  export HOME="/home/devops"
fi
if [ -z "$PYENV_ROOT" ]; then
  export PYENV_ROOT="$HOME/.pyenv"
  export PATH="$PYENV_ROOT/bin:$PATH"
fi
eval "$(pyenv init -)"
eval "$(pyenv virtualenv-init -)"
pyenv activate venv-bai-manager

if [ "$#" -eq 0 ]; then
  exec python -m ai.backend.gateway.server
else
  exec "$@"
fi
```

### Networking

The manager and agent should run in the same local network or different
networks reachable via VPNs, whereas the manager's API service must be exposed to
the public network or another private network that users have access to.

The manager requires access to the etcd, the PostgreSQL database, and the Redis server.

| User-to-Manager TCP Ports | Usage |
|:-------------------------:|-------|
| manager:{80,443}          | Backend.AI API access |

| Manager-to-X TCP Ports | Usage |
|:----------------------:|-------|
| etcd:2379              | etcd API access |
| postgres:5432          | Database access |
| redis:6379             | Redis API access |

The manager must also be able to access TCP ports 6001, 6009, and 30000 to 31000 of the agents in default
configurations.  You can of course change those port numbers and ranges in the configuration.

| Manager-to-Agent TCP Ports | Usage |
|:--------------------------:|-------|
| 6001                       | ZeroMQ-based RPC calls from managers to agents |
| 6009                       | HTTP watcher API |
| 30000-31000                | Port pool for in-container services |


LICENSES
--------

[GNU Lesser General Public License](https://github.com/lablup/backend.ai-manager/blob/master/LICENSE)
[Dependencies](https://github.com/lablup/backend.ai-manager/blob/master/DEPENDENCIES.md)
""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-manager',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.manager': (
            'VERSION',
            'py.typed',
        ),
        'ai.backend.manager.api': (
            'py.typed',
        ),
        'ai.backend.manager.models.alembic': (
            'revision_history/23.09.10.json',
            'revision_history/24.03.11.json',
            'revision_history/24.09.8.json',
            'script.py.mako',
        ),
        'ai.backend.manager.vendor': (
            '__init__.py',
            'bai-icon.svg',
            'graphiql.min.css',
            'graphiql.min.js',
            'react-dom.production.min.js',
            'react.production.min.js',
            'spec-viewer.js',
        ),
    },
    'packages': (
        'ai.backend.manager',
        'ai.backend.manager.actions',
        'ai.backend.manager.actions.monitors',
        'ai.backend.manager.api',
        'ai.backend.manager.cli',
        'ai.backend.manager.client',
        'ai.backend.manager.client.container_registry',
        'ai.backend.manager.container_registry',
        'ai.backend.manager.data.container_registry',
        'ai.backend.manager.data.image',
        'ai.backend.manager.data.resource',
        'ai.backend.manager.data.session',
        'ai.backend.manager.models',
        'ai.backend.manager.models.alembic',
        'ai.backend.manager.models.alembic.versions',
        'ai.backend.manager.models.gql_models',
        'ai.backend.manager.models.gql_models.metric',
        'ai.backend.manager.models.minilang',
        'ai.backend.manager.models.rbac',
        'ai.backend.manager.network',
        'ai.backend.manager.plugin',
        'ai.backend.manager.reporters',
        'ai.backend.manager.scheduler',
        'ai.backend.manager.service',
        'ai.backend.manager.service.container_registry',
        'ai.backend.manager.services',
        'ai.backend.manager.services.agent',
        'ai.backend.manager.services.agent.actions',
        'ai.backend.manager.services.container_registry',
        'ai.backend.manager.services.container_registry.actions',
        'ai.backend.manager.services.domain',
        'ai.backend.manager.services.domain.actions',
        'ai.backend.manager.services.group',
        'ai.backend.manager.services.group.actions',
        'ai.backend.manager.services.image',
        'ai.backend.manager.services.image.actions',
        'ai.backend.manager.services.keypair_resource_policy',
        'ai.backend.manager.services.keypair_resource_policy.actions',
        'ai.backend.manager.services.metric',
        'ai.backend.manager.services.metric.actions',
        'ai.backend.manager.services.metric.processors',
        'ai.backend.manager.services.project_resource_policy',
        'ai.backend.manager.services.project_resource_policy.actions',
        'ai.backend.manager.services.resource_preset',
        'ai.backend.manager.services.resource_preset.actions',
        'ai.backend.manager.services.session',
        'ai.backend.manager.services.session.actions',
        'ai.backend.manager.services.user',
        'ai.backend.manager.services.user.actions',
        'ai.backend.manager.services.user_resource_policy',
        'ai.backend.manager.services.user_resource_policy.actions',
        'ai.backend.manager.services.vfolder',
        'ai.backend.manager.services.vfolder.actions',
        'ai.backend.manager.services.vfolder.processors',
        'ai.backend.manager.services.vfolder.services',
        'ai.backend.manager.sweeper',
        'ai.backend.manager.vendor',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '>=3.13,<3.14',
    'url': 'https://github.com/lablup/backend.ai',
    'version': '25.6.15',
    'zip_safe': False,
})



from __future__ import annotations

import contextlib
import os
import sys
from pathlib import Path
from pprint import pformat
from typing import AsyncIterator, Self

import attrs
import click

from ai.backend.common import redis_helper
from ai.backend.common.config import redis_config_iv
from ai.backend.common.defs import (
    REDIS_IMAGE_DB,
    REDIS_LIVE_DB,
    REDIS_STATISTICS_DB,
    REDIS_STREAM_DB,
    RedisRole,
)
from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
from ai.backend.common.exception import ConfigurationError
from ai.backend.common.types import EtcdRedisConfig, RedisConnectionInfo
from ai.backend.logging import AbstractLogger, LocalLogger, LogLevel

from ..config import LocalConfig, SharedConfig
from ..config import load as load_config


class CLIContext:
    _local_config: LocalConfig | None
    _logger: AbstractLogger

    def __init__(self, config_path: Path, log_level: LogLevel) -> None:
        self.config_path = config_path
        self.log_level = log_level
        self._local_config = None

    @property
    def local_config(self) -> LocalConfig:
        # Lazy-load the configuration only when requested.
        try:
            if self._local_config is None:
                self._local_config = load_config(self.config_path, self.log_level)
        except ConfigurationError as e:
            print(
                "ConfigurationError: Could not read or validate the manager local config:",
                file=sys.stderr,
            )
            print(pformat(e.invalid_data), file=sys.stderr)
            raise click.Abort()
        return self._local_config

    def __enter__(self) -> Self:
        # The "start-server" command is injected by ai.backend.cli from the entrypoint
        # and it has its own multi-process-aware logging initialization.
        # If we duplicate the local logging with it, the process termination may hang.
        click_ctx = click.get_current_context()
        if click_ctx.invoked_subcommand != "start-server":
            logging_config = {
                "level": self.log_level,
                "pkg-ns": {
                    "": LogLevel.WARNING,
                    "ai.backend": self.log_level,
                },
            }
            try:
                # Try getting the logging config but silently fallback to the default if not
                # present (e.g., when `mgr gql show` command used in CI without installation as
                # addressed in #1686).
                with open(os.devnull, "w") as sink, contextlib.redirect_stderr(sink):
                    logging_config = self.local_config["logging"]
            except click.Abort:
                pass
            self._logger = LocalLogger(logging_config)
            self._logger.__enter__()
        return self

    def __exit__(self, *exc_info) -> None:
        click_ctx = click.get_current_context()
        if click_ctx.invoked_subcommand != "start-server":
            self._logger.__exit__()


@contextlib.asynccontextmanager
async def etcd_ctx(cli_ctx: CLIContext) -> AsyncIterator[AsyncEtcd]:
    local_config = cli_ctx.local_config
    creds = None
    if local_config["etcd"]["user"]:
        creds = {
            "user": local_config["etcd"]["user"],
            "password": local_config["etcd"]["password"],
        }
    scope_prefix_map = {
        ConfigScopes.GLOBAL: "",
        # TODO: provide a way to specify other scope prefixes
    }
    etcd = AsyncEtcd(
        local_config["etcd"]["addr"],
        local_config["etcd"]["namespace"],
        scope_prefix_map,
        credentials=creds,
    )
    try:
        yield etcd
    finally:
        await etcd.close()


@contextlib.asynccontextmanager
async def config_ctx(cli_ctx: CLIContext) -> AsyncIterator[SharedConfig]:
    local_config = cli_ctx.local_config
    # scope_prefix_map is created inside ConfigServer
    shared_config = SharedConfig(
        local_config["etcd"]["addr"],
        local_config["etcd"]["user"],
        local_config["etcd"]["password"],
        local_config["etcd"]["namespace"],
    )
    await shared_config.reload()
    raw_redis_config = await shared_config.etcd.get_prefix("config/redis")
    local_config["redis"] = redis_config_iv.check(raw_redis_config)
    try:
        yield shared_config
    finally:
        await shared_config.close()


@attrs.define(auto_attribs=True, frozen=True, slots=True)
class RedisConnectionSet:
    live: RedisConnectionInfo
    stat: RedisConnectionInfo
    image: RedisConnectionInfo
    stream: RedisConnectionInfo


@contextlib.asynccontextmanager
async def redis_ctx(cli_ctx: CLIContext) -> AsyncIterator[RedisConnectionSet]:
    local_config = cli_ctx.local_config
    shared_config = SharedConfig(
        local_config["etcd"]["addr"],
        local_config["etcd"]["user"],
        local_config["etcd"]["password"],
        local_config["etcd"]["namespace"],
    )
    await shared_config.reload()
    raw_redis_config = await shared_config.etcd.get_prefix("config/redis")
    local_config["redis"] = redis_config_iv.check(raw_redis_config)
    redis_config = EtcdRedisConfig.from_dict(local_config["redis"])
    redis_live = redis_helper.get_redis_object(
        redis_config.get_override_config(RedisRole.LIVE),
        name="mgr_cli.live",
        db=REDIS_LIVE_DB,
    )
    redis_stat = redis_helper.get_redis_object(
        redis_config.get_override_config(RedisRole.STATISTICS),
        name="mgr_cli.stat",
        db=REDIS_STATISTICS_DB,
    )
    redis_image = redis_helper.get_redis_object(
        redis_config.get_override_config(RedisRole.IMAGE),
        name="mgr_cli.image",
        db=REDIS_IMAGE_DB,
    )
    redis_stream = redis_helper.get_redis_object(
        redis_config.get_override_config(RedisRole.STREAM),
        name="mgr_cli.stream",
        db=REDIS_STREAM_DB,
    )
    yield RedisConnectionSet(
        live=redis_live,
        stat=redis_stat,
        image=redis_image,
        stream=redis_stream,
    )
    await redis_stream.close()
    await redis_image.close()
    await redis_stat.close()
    await redis_live.close()



import uuid
from dataclasses import dataclass, field
from typing import Any, Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.services.session.base import SessionAction
from ai.backend.manager.types import OptionalState, PartialModifier


@dataclass
class SessionModifier(PartialModifier):
    name: OptionalState[str] = field(default_factory=OptionalState.nop)
    priority: OptionalState[int] = field(default_factory=OptionalState.nop)

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.priority.update_dict(to_update, "priority")
        return to_update


@dataclass
class ModifySessionAction(SessionAction):
    session_id: uuid.UUID
    modifier: SessionModifier

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.session_id)

    @override
    def operation_type(self):
        return "modify"


@dataclass
class ModifySessionActionResult(BaseActionResult):
    # TODO: Add proper type
    result: Any
    session_row: SessionRow

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.session_row.id)



import uuid
from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import Any, Optional, override

from ai.backend.common.types import (
    AccessKey,
    KernelId,
    QuotaScopeID,
    VFolderUsageMode,
)
from ai.backend.manager.actions.action import BaseAction, BaseActionResult
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.models.vfolder import (
    VFolderOperationStatus,
    VFolderOwnershipType,
    VFolderPermission,
)
from ai.backend.manager.types import OptionalState, PartialModifier

from ..types import VFolderBaseInfo, VFolderOwnershipInfo, VFolderUsageInfo


class VFolderAction(BaseAction):
    def entity_type(self):
        return "vfolder"


@dataclass
class CreateVFolderAction(VFolderAction):
    name: str

    keypair_resource_policy: Mapping[str, Any]
    domain_name: str
    group_id_or_name: Optional[str | uuid.UUID]
    folder_host: Optional[str]
    unmanaged_path: Optional[str]
    mount_permission: VFolderPermission
    usage_mode: VFolderUsageMode
    cloneable: bool

    # User identifier
    # TODO: Distinguish between creator and owner
    user_uuid: uuid.UUID
    user_role: UserRole
    creator_email: str

    def entity_id(self) -> Optional[str]:
        return None

    def operation_type(self) -> str:
        return "create"


@dataclass
class CreateVFolderActionResult(BaseActionResult):
    id: uuid.UUID
    name: str
    quota_scope_id: QuotaScopeID
    host: str
    unmanaged_path: Optional[str]
    mount_permission: VFolderPermission
    usage_mode: VFolderUsageMode
    creator_email: str
    ownership_type: VFolderOwnershipType
    user_uuid: Optional[uuid.UUID]
    group_uuid: Optional[uuid.UUID]
    cloneable: bool
    status: VFolderOperationStatus

    def entity_id(self) -> Optional[str]:
        return str(self.id)


@dataclass
class VFolderAttributeModifier(PartialModifier):
    name: OptionalState[str] = field(default_factory=OptionalState.nop)
    cloneable: OptionalState[bool] = field(default_factory=OptionalState.nop)
    mount_permission: OptionalState[VFolderPermission] = field(default_factory=OptionalState.nop)

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.cloneable.update_dict(to_update, "cloneable")
        self.mount_permission.update_dict(to_update, "permission")
        return to_update


@dataclass
class UpdateVFolderAttributeAction(VFolderAction):
    user_uuid: uuid.UUID
    vfolder_uuid: uuid.UUID
    modifier: VFolderAttributeModifier = field(default_factory=VFolderAttributeModifier)

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)

    @override
    def operation_type(self):
        return "update"


@dataclass
class UpdateVFolderAttributeActionResult(BaseActionResult):
    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)


@dataclass
class GetVFolderAction(VFolderAction):
    user_uuid: uuid.UUID
    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)

    @override
    def operation_type(self):
        return "get"


@dataclass
class GetVFolderActionResult(BaseActionResult):
    user_uuid: uuid.UUID
    base_info: VFolderBaseInfo
    ownership_info: VFolderOwnershipInfo
    usage_info: VFolderUsageInfo

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.user_uuid)


@dataclass
class ListVFolderAction(VFolderAction):
    user_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.user_uuid)

    @override
    def operation_type(self):
        return "list"


@dataclass
class ListVFolderActionResult(BaseActionResult):
    user_uuid: uuid.UUID
    vfolders: list[tuple[VFolderBaseInfo, VFolderOwnershipInfo]]

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.user_uuid)


@dataclass
class MoveToTrashVFolderAction(VFolderAction):
    user_uuid: uuid.UUID
    keypair_resource_policy: Mapping[str, Any]

    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)

    @override
    def operation_type(self):
        return "move_to_trash"


@dataclass
class MoveToTrashVFolderActionResult(BaseActionResult):
    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)


@dataclass
class RestoreVFolderFromTrashAction(VFolderAction):
    user_uuid: uuid.UUID

    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)

    @override
    def operation_type(self):
        return "restore"


@dataclass
class RestoreVFolderFromTrashActionResult(BaseActionResult):
    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)


@dataclass
class DeleteForeverVFolderAction(VFolderAction):
    user_uuid: uuid.UUID

    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)

    @override
    def operation_type(self):
        return "delete_forever"


@dataclass
class DeleteForeverVFolderActionResult(BaseActionResult):
    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)


@dataclass
class ForceDeleteVFolderAction(VFolderAction):
    """
    This action transits the state of vfolder from ready to delete-forever directly.
    """

    user_uuid: uuid.UUID

    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)

    @override
    def operation_type(self):
        return "force_delete"


@dataclass
class ForceDeleteVFolderActionResult(BaseActionResult):
    vfolder_uuid: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)


@dataclass
class CloneVFolderAction(VFolderAction):
    requester_user_uuid: uuid.UUID

    source_vfolder_uuid: uuid.UUID
    target_name: str
    target_host: Optional[str]
    cloneable: bool
    usage_mode: VFolderUsageMode
    mount_permission: VFolderPermission

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.source_vfolder_uuid)

    @override
    def operation_type(self):
        return "clone"


@dataclass
class CloneVFolderActionResult(BaseActionResult):
    vfolder_uuid: uuid.UUID

    target_vfolder_id: uuid.UUID
    target_vfolder_name: str
    target_vfolder_host: str
    usage_mode: VFolderUsageMode
    mount_permission: VFolderPermission
    creator_email: str
    ownership_type: VFolderOwnershipType
    owner_user_uuid: Optional[uuid.UUID]
    owner_group_uuid: Optional[uuid.UUID]
    cloneable: bool
    bgtask_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_uuid)


@dataclass
class GetTaskLogsAction(VFolderAction):
    user_id: uuid.UUID
    domain_name: str
    user_role: UserRole
    kernel_id: KernelId
    owner_access_key: AccessKey

    # TODO: Remove this.
    request: Any

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self):
        return "get_task_logs"


@dataclass
class GetTaskLogsActionResult(BaseActionResult):
    # TODO: Add proper type
    response: Any
    # TODO: Replace this with VFolderData
    vfolder_data: Any

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.vfolder_data["id"])



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.user.actions.base import UserAction
from ai.backend.manager.services.user.type import UserCreator, UserData


@dataclass
class CreateUserAction(UserAction):
    input: UserCreator

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self) -> str:
        return "create"


@dataclass
class CreateUserActionResult(BaseActionResult):
    data: Optional[UserData]
    success: bool

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.data.id) if self.data else None



from dataclasses import dataclass
from typing import Any, Optional, override

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.services.session.base import SessionAction


@dataclass
class StartServiceAction(SessionAction):
    session_name: str
    access_key: AccessKey
    service: str
    login_session_token: Any
    port: Optional[int]
    arguments: Optional[str]  # json_string
    envs: Optional[str]  # json_string

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self):
        return "start_service"


@dataclass
class StartServiceActionResult(BaseActionResult):
    # TODO: Add proper type
    result: Any
    session_row: SessionRow
    token: str
    wsproxy_addr: str

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.session_row.id)



from dataclasses import dataclass
from typing import Any, Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.user.actions.base import UserAction


# TODO: Batch action으로 만든 후 entity_id엔 모든 user_id를 넣어야 함.
@dataclass
class AdminMonthStatsAction(UserAction):
    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self):
        return "admin_month_stats"


@dataclass
class AdminMonthStatsActionResult(BaseActionResult):
    stats: list[Any]

    @override
    def entity_id(self) -> Optional[str]:
        return None



"""Add AuditLogs table

Revision ID: 683ca0a32f41
Revises: 70e0533aa49a
Create Date: 2025-02-14 10:56:10.191119

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "683ca0a32f41"
down_revision = "70e0533aa49a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "audit_logs",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("entity_type", sa.String, nullable=False),
        sa.Column("operation", sa.String, nullable=False),
        sa.Column("entity_id", sa.String, nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "request_id",
            GUID,
            nullable=False,
        ),
        sa.Column("description", sa.String, nullable=False),
        sa.Column("duration", sa.Interval, nullable=False),
        sa.Column("status", sa.VARCHAR(length=64), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_audit_logs")),
    )
    op.create_index(op.f("ix_audit_logs_created_at"), "audit_logs", ["created_at"], unique=False)
    op.create_index(op.f("ix_audit_logs_entity_type"), "audit_logs", ["entity_type"], unique=False)
    op.create_index(op.f("ix_audit_logs_operation"), "audit_logs", ["operation"], unique=False)
    op.create_index(op.f("ix_audit_logs_entity_id"), "audit_logs", ["entity_id"], unique=False)


def downgrade() -> None:
    op.drop_index(op.f("ix_audit_logs_created_at"), table_name="audit_logs")
    op.drop_index(op.f("ix_audit_logs_entity_type"), table_name="audit_logs")
    op.drop_index(op.f("ix_audit_logs_operation"), table_name="audit_logs")
    op.drop_index(op.f("ix_audit_logs_entity_id"), table_name="audit_logs")
    op.drop_table("audit_logs")



from aiohttp import web

from ai.backend.common.exception import BackendError


class VFolderServiceException(Exception):
    pass


class VFolderNotFound(VFolderServiceException):
    pass


class VFolderCreationFailure(VFolderServiceException):
    pass


class VFolderAlreadyExists(BackendError, web.HTTPConflict):
    pass


class InvalidParameter(VFolderServiceException):
    pass


class InsufficientPrivilege(VFolderServiceException):
    pass


class Forbidden(InvalidParameter):
    pass


class ObjectNotFound(VFolderServiceException):
    pass


class ProjectNotFound(VFolderServiceException):
    pass


class InternalServerError(VFolderServiceException):
    pass


class ModelServiceDependencyNotCleared(VFolderServiceException):
    pass


class TooManyVFoldersFound(VFolderServiceException):
    pass


class VFolderFilterStatusNotAvailable(VFolderServiceException):
    pass


class VFolderFilterStatusFailed(VFolderServiceException):
    pass



"""change_main_access_key_ondelete_to_set_null

Revision ID: c5ed277b7f7b
Revises: d3f8c74bf148
Create Date: 2024-01-11 15:37:24.097596

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "c5ed277b7f7b"
down_revision = "d3f8c74bf148"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint("fk_users_main_access_key_keypairs", "users", type_="foreignkey")
    op.create_foreign_key(
        op.f("fk_users_main_access_key_keypairs"),
        "users",
        "keypairs",
        ["main_access_key"],
        ["access_key"],
        ondelete="SET NULL",
    )


def downgrade():
    op.drop_constraint(op.f("fk_users_main_access_key_keypairs"), "users", type_="foreignkey")
    op.create_foreign_key(
        "fk_users_main_access_key_keypairs",
        "users",
        "keypairs",
        ["main_access_key"],
        ["access_key"],
        ondelete="RESTRICT",
    )



"""replace_kernelrole_to_sessiontypes

Revision ID: 3596bc12ec09
Revises: c4b7ec740b36
Create Date: 2023-10-04 16:43:46.281383

"""

import enum
import uuid
from typing import cast

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import text

from ai.backend.manager.models.base import GUID, mapper_registry

# revision identifiers, used by Alembic.
revision = "3596bc12ec09"
down_revision = "c4b7ec740b36"
branch_labels = None
depends_on = None

ENUM_CLS = "sessiontypes"

PAGE_SIZE = 100


class KernelRole(enum.Enum):
    INFERENCE = "INFERENCE"
    COMPUTE = "COMPUTE"
    SYSTEM = "SYSTEM"


kernelrole_choices = list(map(lambda v: v.name, KernelRole))
kernelrole = postgresql.ENUM(*kernelrole_choices, name="kernelrole")


class OldSessionTypes(enum.StrEnum):
    INTERACTIVE = "interactive"
    BATCH = "batch"
    INFERENCE = "inference"


def upgrade():
    connection = op.get_bind()

    # Relax the sessions.session_type from enum to varchar(64).
    connection.execute(
        text(
            "ALTER TABLE sessions ALTER COLUMN session_type TYPE varchar(64) USING session_type::text;"
        )
    )
    connection.execute(
        text("ALTER TABLE sessions ALTER COLUMN session_type SET DEFAULT 'INTERACTIVE';")
    )

    # Relax the kernels.session_type from enum to varchar(64).
    connection.execute(
        text(
            "ALTER TABLE kernels ALTER COLUMN session_type TYPE varchar(64) USING session_type::text;"
        )
    )
    connection.execute(
        text("ALTER TABLE kernels ALTER COLUMN session_type SET DEFAULT 'INTERACTIVE';")
    )
    # Relax the kernels.role from enum to varchar(64).
    connection.execute(
        text("ALTER TABLE kernels ALTER COLUMN role TYPE varchar(64) USING role::text;")
    )
    connection.execute(text("ALTER TABLE kernels ALTER COLUMN role SET DEFAULT 'COMPUTE';"))

    # Drop enum types
    connection.execute(text("DROP TYPE IF EXISTS sessiontypes;"))
    connection.execute(text("DROP TYPE IF EXISTS kernelrole;"))

    # Update `sessions.session_type` column
    kernels = sa.Table(
        "kernels",
        mapper_registry.metadata,
        sa.Column("id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")),
        sa.Column("session_id", GUID),
        sa.Column(
            "role",
            postgresql.ENUM("INFERENCE", "COMPUTE", "SYSTEM", name="kernelrole"),
            default=KernelRole.COMPUTE,
            server_default=KernelRole.COMPUTE.name,
        ),
        sa.Column(
            "session_type",
            sa.VARCHAR,
            index=True,
            nullable=False,
            default="interactive",
            server_default="interactive",
        ),
        extend_existing=True,
    )

    sessions = sa.Table(
        "sessions",
        mapper_registry.metadata,
        sa.Column("id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")),
        sa.Column(
            "session_type",
            sa.VARCHAR,
            index=True,
            nullable=False,
            default="interactive",
            server_default="interactive",
        ),
        extend_existing=True,
    )
    while True:
        _sid_query = (
            sa.select(kernels.c.session_id).where(kernels.c.role == "SYSTEM").limit(PAGE_SIZE)
        )
        session_ids = cast(list[uuid.UUID], connection.scalars(_sid_query).all())
        _session_query = (
            sa.update(sessions)
            .values({"session_type": "SYSTEM"})
            .where(sessions.c.id.in_(session_ids))
        )
        _kernel_query = (
            sa.update(kernels)
            .values({"session_type": "SYSTEM", "role": "COMPUTE"})
            .where(kernels.c.session_id.in_(session_ids))
        )
        connection.execute(_session_query)
        result = connection.execute(_kernel_query)

        if result.rowcount == 0:
            break

    op.drop_column("kernels", "role")


def downgrade():
    connection = op.get_bind()

    kernel_role_values = ["INFERENCE", "COMPUTE", "SYSTEM"]
    KernelRoleType = postgresql.ENUM(*kernel_role_values, name="kernelrole")
    KernelRoleType.create(connection)
    op.add_column(
        "kernels",
        sa.Column(
            "role",
            KernelRoleType,
            autoincrement=False,
            nullable=True,
            server_default=KernelRole.COMPUTE.name,
        ),
    )

    kernels = sa.Table(
        "kernels",
        mapper_registry.metadata,
        sa.Column("id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")),
        sa.Column(
            "role",
            KernelRoleType,
            default=KernelRole.COMPUTE,
            server_default=KernelRole.COMPUTE.name,
        ),
        sa.Column(
            "session_type",
            sa.VARCHAR,
            index=True,
            nullable=False,
            default="interactive",
            server_default="interactive",
        ),
        extend_existing=True,
    )

    sessions = sa.Table(
        "sessions",
        mapper_registry.metadata,
        sa.Column("id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")),
        sa.Column(
            "session_type",
            sa.VARCHAR,
            index=True,
            nullable=False,
            default="interactive",
            server_default="interactive",
        ),
        extend_existing=True,
    )

    # replace session_type.system role
    while True:
        _sid_query = (
            sa.select(kernels.c.session_id)
            .where(kernels.c.session_type == "SYSTEM")
            .limit(PAGE_SIZE)
        )
        session_ids = cast(list[uuid.UUID], connection.scalars(_sid_query).all())
        _session_query = (
            sa.update(sessions)
            .values({"session_type": "INTERACTIVE"})
            .where(sessions.c.id.in_(session_ids))
        )
        _kernel_query = (
            sa.update(kernels)
            .values({"session_type": "INTERACTIVE", "role": "SYSTEM"})
            .where(kernels.c.session_id.in_(session_ids))
        )
        connection.execute(_session_query)
        result = connection.execute(_kernel_query)
        if result.rowcount == 0:
            break

    op.alter_column("kernels", column_name="role", nullable=False)

    connection.execute(
        text(
            "CREATE TYPE sessiontypes AS ENUM (%s)"
            % (",".join(f"'{choice.name}'" for choice in OldSessionTypes))
        )
    )
    # Revert sessions.session_type to enum
    connection.execute(text("ALTER TABLE sessions ALTER COLUMN session_type DROP DEFAULT;"))
    connection.execute(
        text(
            "ALTER TABLE sessions ALTER COLUMN session_type TYPE sessiontypes "
            "USING session_type::sessiontypes;"
        )
    )
    connection.execute(
        text("ALTER TABLE sessions ALTER COLUMN session_type SET DEFAULT 'INTERACTIVE';")
    )

    # Revert kernels.session_type to enum
    connection.execute(text("ALTER TABLE kernels ALTER COLUMN session_type DROP DEFAULT;"))
    connection.execute(
        text(
            "ALTER TABLE kernels ALTER COLUMN session_type TYPE sessiontypes "
            "USING session_type::sessiontypes;"
        )
    )
    connection.execute(
        text("ALTER TABLE kernels ALTER COLUMN session_type SET DEFAULT 'INTERACTIVE';")
    )



"""remove unique constraint of endpoint.url

Revision ID: 8b2ec7e3d22a
Revises: c5ed277b7f7b
Create Date: 2024-01-11 11:02:39.253292

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "8b2ec7e3d22a"
down_revision = "c5ed277b7f7b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("uq_endpoints_url", "endpoints", type_="unique")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint("uq_endpoints_url", "endpoints", ["url"])
    # ### end Alembic commands ###



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    AsyncContextManager,
    Awaitable,
    Callable,
    Iterable,
    Mapping,
    Tuple,
)

import aiohttp_cors
from aiohttp import web
from aiohttp.typedefs import Middleware
from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from .context import RootContext


WebRequestHandler: TypeAlias = Callable[
    [web.Request],
    Awaitable[web.StreamResponse],
]
WebMiddleware: TypeAlias = Middleware

CORSOptions: TypeAlias = Mapping[str, aiohttp_cors.ResourceOptions]
AppCreator: TypeAlias = Callable[
    [CORSOptions],
    Tuple[web.Application, Iterable[WebMiddleware]],
]

CleanupContext: TypeAlias = Callable[["RootContext"], AsyncContextManager[None]]



from __future__ import annotations

import logging
from collections.abc import (
    Container,
    Mapping,
    Sequence,
)
from decimal import Decimal
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Optional,
    Self,
    cast,
)
from uuid import UUID, uuid4

import sqlalchemy as sa
import trafaret as t
import yarl
from redis.asyncio import Redis
from redis.asyncio.client import Pipeline
from ruamel.yaml import YAML
from ruamel.yaml.error import YAMLError
from sqlalchemy import CheckConstraint
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncSession
from sqlalchemy.orm import contains_eager, foreign, relationship, selectinload
from sqlalchemy.orm.exc import NoResultFound

from ai.backend.common import msgpack, redis_helper
from ai.backend.common.config import model_definition_iv
from ai.backend.common.types import (
    AccessKey,
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
    ClusterMode,
    EndpointId,
    RedisConnectionInfo,
    RuntimeVariant,
    SessionTypes,
    VFolderID,
    VFolderMount,
    VFolderUsageMode,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.defs import DEFAULT_CHUNK_SIZE
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.types import MountOptionModel, UserScope

from ..api.exceptions import (
    InvalidAPIParameters,
    ObjectNotFound,
    ServiceUnavailable,
)
from .base import (
    GUID,
    Base,
    DecimalType,
    EndpointIDColumn,
    EnumValueType,
    IDColumn,
    ResourceSlotColumn,
    StrEnumType,
    StructuredJSONObjectListColumn,
    URLColumn,
)
from .image import ImageRow
from .routing import RouteStatus
from .scaling_group import scaling_groups
from .user import UserRow
from .vfolder import VFolderRow, prepare_vfolder_mounts

if TYPE_CHECKING:
    from ai.backend.manager.config import SharedConfig

    from .gql import GraphQueryContext

__all__ = (
    "EndpointRow",
    "EndpointLifecycle",
    "ModelServicePredicateChecker",
    "EndpointStatistics",
    "EndpointTokenRow",
    "EndpointAutoScalingRuleRow",
)


log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore


class EndpointLifecycle(Enum):
    CREATED = "created"
    DESTROYING = "destroying"
    DESTROYED = "destroyed"


class EndpointRow(Base):
    __tablename__ = "endpoints"

    __table_args__ = (
        CheckConstraint(
            sa.or_(
                sa.column("lifecycle_stage") == EndpointLifecycle.DESTROYED.value,
                sa.column("image").isnot(None),
            ),
            name="ck_image_required_unless_destroyed",
        ),
    )

    id = EndpointIDColumn()
    name = sa.Column("name", sa.String(length=512), nullable=False)
    created_user = sa.Column("created_user", GUID, nullable=False)
    session_owner = sa.Column("session_owner", GUID, nullable=False)
    # minus session count means this endpoint is requested for removal
    replicas = sa.Column("replicas", sa.Integer, nullable=False, default=0, server_default="0")
    image = sa.Column("image", GUID)
    model = sa.Column(
        "model",
        GUID,
        sa.ForeignKey("vfolders.id", ondelete="SET NULL"),
        nullable=True,
    )
    model_mount_destination = sa.Column(
        "model_mount_destination",
        sa.String(length=1024),
        nullable=False,
        default="/models",
        server_default="/models",
    )
    domain = sa.Column(
        "domain",
        sa.String(length=64),
        sa.ForeignKey("domains.name", ondelete="RESTRICT"),
        nullable=False,
    )
    project = sa.Column(
        "project",
        GUID,
        sa.ForeignKey("groups.id", ondelete="RESTRICT"),
        nullable=False,
    )
    resource_group = sa.Column(
        "resource_group",
        sa.ForeignKey("scaling_groups.name", ondelete="RESTRICT"),
        index=True,
        nullable=False,
    )
    lifecycle_stage = sa.Column(
        "lifecycle_stage",
        EnumValueType(EndpointLifecycle),
        nullable=False,
        default=EndpointLifecycle.CREATED,
    )
    tag = sa.Column("tag", sa.String(length=64), nullable=True)
    startup_command = sa.Column("startup_command", sa.Text, nullable=True)
    bootstrap_script = sa.Column("bootstrap_script", sa.String(length=16 * 1024), nullable=True)
    callback_url = sa.Column("callback_url", URLColumn, nullable=True, default=sa.null())
    environ = sa.Column("environ", pgsql.JSONB(), nullable=True, default={})
    open_to_public = sa.Column("open_to_public", sa.Boolean, default=False)
    runtime_variant = sa.Column(
        "runtime_variant",
        StrEnumType(RuntimeVariant),
        nullable=False,
        default=RuntimeVariant.CUSTOM,
    )

    model_definition_path = sa.Column("model_definition_path", sa.String(length=128), nullable=True)

    resource_slots = sa.Column("resource_slots", ResourceSlotColumn(), nullable=False)
    url = sa.Column("url", sa.String(length=1024))
    resource_opts = sa.Column("resource_opts", pgsql.JSONB(), nullable=True, default={})
    cluster_mode = sa.Column(
        "cluster_mode",
        sa.String(length=16),
        nullable=False,
        default=ClusterMode.SINGLE_NODE,
        server_default=ClusterMode.SINGLE_NODE.name,
    )
    cluster_size = sa.Column(
        "cluster_size", sa.Integer, nullable=False, default=1, server_default="1"
    )

    extra_mounts = sa.Column(
        "extra_mounts",
        StructuredJSONObjectListColumn(VFolderMount),
        nullable=False,
        default=[],
        server_default="[]",
    )

    retries = sa.Column("retries", sa.Integer, nullable=False, default=0, server_default="0")
    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.text("now()"),
        nullable=True,
    )
    destroyed_at = sa.Column(
        "destroyed_at",
        sa.DateTime(timezone=True),
        nullable=True,
    )

    routings = relationship("RoutingRow", back_populates="endpoint_row")
    tokens = relationship(
        "EndpointTokenRow",
        back_populates="endpoint_row",
        primaryjoin="foreign(EndpointTokenRow.endpoint) == EndpointRow.id",
    )
    endpoint_auto_scaling_rules = relationship(
        "EndpointAutoScalingRuleRow", back_populates="endpoint_row"
    )
    image_row = relationship(
        "ImageRow",
        primaryjoin=lambda: foreign(EndpointRow.image) == ImageRow.id,
        foreign_keys=[image],
        back_populates="endpoints",
    )

    model_row = relationship("VFolderRow", back_populates="endpoints")
    created_user_row = relationship(
        "UserRow",
        back_populates="created_endpoints",
        foreign_keys=[created_user],
        primaryjoin=lambda: foreign(EndpointRow.created_user) == UserRow.uuid,
    )
    session_owner_row = relationship(
        "UserRow",
        back_populates="owned_endpoints",
        foreign_keys=[session_owner],
        primaryjoin=lambda: foreign(EndpointRow.session_owner) == UserRow.uuid,
    )

    def __init__(
        self,
        name: str,
        model_definition_path: str | None,
        created_user: UUID,
        session_owner: UUID,
        replicas: int,
        image: ImageRow,
        model: UUID,
        domain: str,
        project: UUID,
        resource_group: str,
        resource_slots: Mapping[str, Any],
        cluster_mode: ClusterMode,
        cluster_size: int,
        extra_mounts: Sequence[VFolderMount],
        runtime_variant: RuntimeVariant,
        *,
        model_mount_destination: Optional[str] = None,
        tag: Optional[str] = None,
        startup_command: Optional[str] = None,
        bootstrap_script: Optional[str] = None,
        callback_url: Optional[yarl.URL] = None,
        environ: Optional[Mapping[str, Any]] = None,
        resource_opts: Optional[Mapping[str, Any]] = None,
        open_to_public: bool = False,
    ):
        self.id = uuid4()
        self.name = name
        self.model_definition_path = model_definition_path
        self.created_user = created_user
        self.session_owner = session_owner
        self.replicas = replicas
        self.image = image.id
        self.model = model
        self.domain = domain
        self.project = project
        self.resource_group = resource_group
        self.resource_slots = resource_slots
        self.cluster_mode = cluster_mode
        self.cluster_size = cluster_size
        self.extra_mounts = extra_mounts
        self.model_mount_destination = model_mount_destination or "/models"
        self.tag = tag
        self.startup_command = startup_command
        self.bootstrap_script = bootstrap_script
        self.callback_url = callback_url
        self.environ = environ
        self.resource_opts = resource_opts
        self.open_to_public = open_to_public
        self.runtime_variant = runtime_variant

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        endpoint_id: UUID,
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_routes: bool = False,
        load_tokens: bool = False,
        load_image: bool = False,
        load_created_user: bool = False,
        load_session_owner: bool = False,
        load_model: bool = False,
    ) -> Self:
        """
        :raises: sqlalchemy.orm.exc.NoResultFound
        """
        query = sa.select(EndpointRow).filter(EndpointRow.id == endpoint_id)
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(
                selectinload(EndpointRow.image_row).selectinload(ImageRow.aliases)
            )
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if load_model:
            query = query.options(selectinload(EndpointRow.model_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        row = result.scalar()
        if row is None:
            raise NoResultFound
        return row

    @classmethod
    async def list(
        cls,
        session: AsyncSession,
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_routes: bool = False,
        load_image: bool = False,
        load_tokens: bool = False,
        load_created_user: bool = False,
        load_session_owner: bool = False,
        status_filter: Container[EndpointLifecycle] = frozenset([EndpointLifecycle.CREATED]),
    ) -> list[Self]:
        query = (
            sa.select(EndpointRow)
            .order_by(sa.desc(EndpointRow.created_at))
            .filter(EndpointRow.lifecycle_stage.in_(status_filter))
        )
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(selectinload(EndpointRow.image_row))
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def batch_load(
        cls,
        session: AsyncSession,
        endpoint_ids: Sequence[EndpointId],
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_routes: bool = False,
        load_image: bool = False,
        load_tokens: bool = False,
        load_created_user: bool = False,
        load_session_owner: bool = False,
        status_filter: Container[EndpointLifecycle] = frozenset([EndpointLifecycle.CREATED]),
    ) -> Sequence[Self]:
        query = (
            sa.select(EndpointRow)
            .order_by(sa.desc(EndpointRow.created_at))
            .filter(
                EndpointRow.lifecycle_stage.in_(status_filter) & EndpointRow.id.in_(endpoint_ids)
            )
        )
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(selectinload(EndpointRow.image_row))
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def list_by_model(
        cls,
        session: AsyncSession,
        model_id: UUID,
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_routes: bool = False,
        load_image: bool = False,
        load_tokens: bool = False,
        load_created_user: bool = False,
        load_session_owner: bool = False,
        status_filter: Container[EndpointLifecycle] = frozenset([EndpointLifecycle.CREATED]),
    ) -> Sequence[Self]:
        query = (
            sa.select(EndpointRow)
            .order_by(sa.desc(EndpointRow.created_at))
            .filter(
                EndpointRow.lifecycle_stage.in_(status_filter) & (EndpointRow.model == model_id)
            )
        )
        if load_routes:
            query = query.options(selectinload(EndpointRow.routings))
        if load_tokens:
            query = query.options(selectinload(EndpointRow.tokens))
        if load_image:
            query = query.options(selectinload(EndpointRow.image_row))
        if load_created_user:
            query = query.options(selectinload(EndpointRow.created_user_row))
        if load_session_owner:
            query = query.options(selectinload(EndpointRow.session_owner_row))
        if project:
            query = query.filter(EndpointRow.project == project)
        if domain:
            query = query.filter(EndpointRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()

    async def create_auto_scaling_rule(
        self,
        session: AsyncSession,
        metric_source: AutoScalingMetricSource,
        metric_name: str,
        threshold: Decimal,
        comparator: AutoScalingMetricComparator,
        step_size: int,
        cooldown_seconds: int = 300,
        min_replicas: int | None = None,
        max_replicas: int | None = None,
    ) -> EndpointAutoScalingRuleRow:
        row = EndpointAutoScalingRuleRow(
            id=uuid4(),
            endpoint=self.id,
            metric_source=metric_source,
            metric_name=metric_name,
            threshold=threshold,
            comparator=comparator,
            step_size=step_size,
            cooldown_seconds=cooldown_seconds,
            min_replicas=min_replicas,
            max_replicas=max_replicas,
        )
        session.add(row)
        return row

    @property
    def terminatable_route_statuses(self) -> set[RouteStatus]:
        if self.lifecycle_stage == EndpointLifecycle.DESTROYING:
            return {
                RouteStatus.PROVISIONING,
                RouteStatus.HEALTHY,
                RouteStatus.UNHEALTHY,
                RouteStatus.FAILED_TO_START,
            }
        else:
            return {
                RouteStatus.HEALTHY,
                RouteStatus.UNHEALTHY,
                RouteStatus.FAILED_TO_START,
            }

    @staticmethod
    async def delegate_endpoint_ownership(
        db_session: AsyncSession,
        owner_user_uuid: UUID,
        target_user_uuid: UUID,
        target_access_key: AccessKey,
    ) -> None:
        from .routing import RoutingRow
        from .session import KernelLoadingStrategy, SessionRow

        endpoint_rows = await EndpointRow.list(
            db_session,
            user_uuid=owner_user_uuid,
            load_session_owner=True,
            load_routes=True,
            load_tokens=True,
        )
        session_ids: List[UUID] = []
        for row in endpoint_rows:
            row.session_owner = target_user_uuid
            for token_row in cast(list[EndpointTokenRow], row.tokens):
                token_row.delegate_ownership(target_user_uuid)
            for routing_row in cast(list[RoutingRow], row.routings):
                routing_row.delegate_ownership(target_user_uuid)
                session_ids.append(routing_row.session)
        session_rows = await SessionRow.list_sessions(
            db_session, session_ids, kernel_loading_strategy=KernelLoadingStrategy.ALL_KERNELS
        )
        for session_row in session_rows:
            session_row.delegate_ownership(target_user_uuid, target_access_key)


class EndpointTokenRow(Base):
    __tablename__ = "endpoint_tokens"

    id = IDColumn()
    token = sa.Column("token", sa.String(), nullable=False)
    endpoint = sa.Column("endpoint", GUID, nullable=True)
    session_owner = sa.Column("session_owner", GUID, nullable=False)
    domain = sa.Column(
        "domain",
        sa.String(length=64),
        sa.ForeignKey("domains.name", ondelete="CASCADE"),
        nullable=False,
    )
    project = sa.Column(
        "project",
        GUID,
        sa.ForeignKey("groups.id", ondelete="CASCADE"),
        nullable=False,
    )
    created_at = sa.Column(
        "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
    )

    endpoint_row = relationship(
        "EndpointRow",
        back_populates="tokens",
        foreign_keys=[endpoint],
        primaryjoin=lambda: foreign(EndpointTokenRow.endpoint) == EndpointRow.id,
    )

    def __init__(
        self,
        id: UUID,
        token: str,
        endpoint: UUID,
        domain: str,
        project: UUID,
        session_owner: UUID,
    ) -> None:
        self.id = id
        self.token = token
        self.endpoint = endpoint
        self.domain = domain
        self.project = project
        self.session_owner = session_owner

    @classmethod
    async def list(
        cls,
        session: AsyncSession,
        endpoint_id: UUID,
        *,
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_endpoint=False,
    ) -> Sequence[Self]:
        query = (
            sa.select(EndpointTokenRow)
            .filter(EndpointTokenRow.endpoint == endpoint_id)
            .order_by(sa.desc(EndpointTokenRow.created_at))
        )
        if load_endpoint:
            query = query.options(selectinload(EndpointTokenRow.tokens))
        if project:
            query = query.filter(EndpointTokenRow.project == project)
        if domain:
            query = query.filter(EndpointTokenRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointTokenRow.session_owner == user_uuid)
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        token: str,
        *,
        domain: Optional[str] = None,
        project: Optional[UUID] = None,
        user_uuid: Optional[UUID] = None,
        load_endpoint=False,
    ) -> Self:
        query = sa.select(EndpointTokenRow).filter(EndpointTokenRow.token == token)
        if load_endpoint:
            query = query.options(selectinload(EndpointTokenRow.tokens))
        if project:
            query = query.filter(EndpointTokenRow.project == project)
        if domain:
            query = query.filter(EndpointTokenRow.domain == domain)
        if user_uuid:
            query = query.filter(EndpointTokenRow.session_owner == user_uuid)
        result = await session.execute(query)
        row = result.scalar()
        if not row:
            raise NoResultFound
        return row

    def delegate_ownership(self, user_uuid: UUID) -> None:
        self.session_owner = user_uuid


class EndpointAutoScalingRuleRow(Base):
    __tablename__ = "endpoint_auto_scaling_rules"

    id = IDColumn()
    metric_source = sa.Column(
        "metric_source", StrEnumType(AutoScalingMetricSource, use_name=False), nullable=False
    )
    metric_name = sa.Column("metric_name", sa.Text(), nullable=False)
    threshold = sa.Column("threshold", DecimalType(), nullable=False)
    comparator = sa.Column(
        "comparator", StrEnumType(AutoScalingMetricComparator, use_name=False), nullable=False
    )
    step_size = sa.Column("step_size", sa.Integer(), nullable=False)
    cooldown_seconds = sa.Column("cooldown_seconds", sa.Integer(), nullable=False, default=300)

    min_replicas = sa.Column("min_replicas", sa.Integer(), nullable=True)
    max_replicas = sa.Column("max_replicas", sa.Integer(), nullable=True)

    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.text("now()"),
        nullable=True,
    )
    last_triggered_at = sa.Column(
        "last_triggered_at",
        sa.DateTime(timezone=True),
        nullable=True,
    )

    endpoint = sa.Column(
        "endpoint",
        GUID,
        sa.ForeignKey("endpoints.id", ondelete="CASCADE"),
        nullable=False,
    )

    endpoint_row = relationship(
        "EndpointRow", back_populates="endpoint_auto_scaling_rules", lazy="joined"
    )

    @classmethod
    async def list(
        cls,
        session: AsyncSession,
        endpoint_status_filter: Container[EndpointLifecycle] = frozenset([
            EndpointLifecycle.CREATED
        ]),
    ) -> Sequence[Self]:
        query = sa.select(EndpointAutoScalingRuleRow)
        if endpoint_status_filter:
            query = (
                query.join(EndpointAutoScalingRuleRow.endpoint_row)
                .filter(EndpointRow.lifecycle_stage.in_(endpoint_status_filter))
                .options(contains_eager(EndpointAutoScalingRuleRow.endpoint_row))
            )
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def get(
        cls, session: AsyncSession, id: UUID, load_endpoint=False
    ) -> "EndpointAutoScalingRuleRow":
        query = sa.select(EndpointAutoScalingRuleRow).filter(EndpointAutoScalingRuleRow.id == id)
        if load_endpoint:
            query = query.options(selectinload(EndpointAutoScalingRuleRow.endpoint_row))
        result = await session.execute(query)
        row = result.scalar()
        if not row:
            raise ObjectNotFound(object_name="Endpoint Autoscaling Rule")
        return row

    async def remove_rule(
        self,
        session: AsyncSession,
    ) -> None:
        await session.delete(self)


class ModelServicePredicateChecker:
    @staticmethod
    async def check_scaling_group(
        conn: AsyncConnection,
        scaling_group: str,
        owner_access_key: AccessKey,
        target_domain: str,
        target_project: str | UUID,
    ) -> str:
        """
        Wrapper of `registry.check_scaling_group()` with additional guards flavored for
        model service included
        """
        from ai.backend.manager.registry import check_scaling_group

        checked_scaling_group = await check_scaling_group(
            conn,
            scaling_group,
            SessionTypes.INFERENCE,
            owner_access_key,
            target_domain,
            target_project,
        )

        query = (
            sa.select([scaling_groups.c.wsproxy_addr, scaling_groups.c.wsproxy_api_token])
            .select_from(scaling_groups)
            .where((scaling_groups.c.name == checked_scaling_group))
        )

        result = await conn.execute(query)
        sgroup = result.first()
        wsproxy_addr = sgroup["wsproxy_addr"]
        if not wsproxy_addr:
            raise ServiceUnavailable("No coordinator configured for this resource group")

        if not sgroup["wsproxy_api_token"]:
            raise ServiceUnavailable("Scaling group not ready to start model service")

        return checked_scaling_group

    @staticmethod
    async def check_extra_mounts(
        conn: AsyncConnection,
        shared_config: "SharedConfig",
        storage_manager: StorageSessionManager,
        model_id: UUID,
        model_mount_destination: str,
        extra_mounts: dict[UUID, MountOptionModel],
        user_scope: UserScope,
        resource_policy: dict[str, Any],
    ) -> Sequence[VFolderMount]:
        """
        check if user is allowed to access every folders eagering to mount (other than model VFolder)
        on general session creation lifecycle this check will be completed by `enqueue_session()` function,
        which is not covered by the validation procedure (`create_session(dry_run=True)` call at the bottom part of `create()` API)
        so we have to manually cover this part here.
        """

        if model_id in extra_mounts:
            raise InvalidAPIParameters(
                "Same VFolder appears on both model specification and VFolder mount"
            )

        requested_mounts = [*extra_mounts.keys()]
        requested_mount_map: dict[str | UUID, str] = {
            folder_id: options.mount_destination
            for folder_id, options in extra_mounts.items()
            if options.mount_destination
        }
        requested_mount_options: dict[str | UUID, Any] = {
            folder_id: {
                "type": options.type,
                "permission": options.permission,
            }
            for folder_id, options in extra_mounts.items()
        }
        log.debug(
            "requested mounts: {}, mount_map: {}, mount_options: {}",
            requested_mounts,
            requested_mount_map,
            requested_mount_options,
        )
        allowed_vfolder_types = await shared_config.get_vfolder_types()
        vfolder_mounts = await prepare_vfolder_mounts(
            conn,
            storage_manager,
            allowed_vfolder_types,
            user_scope,
            resource_policy,
            requested_mounts,
            requested_mount_map,
            requested_mount_options,
        )

        for vfolder in vfolder_mounts:
            if vfolder.kernel_path == model_mount_destination:
                raise InvalidAPIParameters(
                    "extra_mounts.mount_destination conflicts with model_mount_destination config. Make sure not to shadow value defined at model_mount_destination as a mount destination of extra VFolders."
                )
            if vfolder.usage_mode == VFolderUsageMode.MODEL:
                raise InvalidAPIParameters(
                    "MODEL type VFolders cannot be added as a part of extra_mounts folder"
                )

        return vfolder_mounts

    @staticmethod
    async def _listdir(
        storage_manager: StorageSessionManager,
        proxy_name: str,
        volume_name: str,
        vfid: VFolderID,
        relpath: str,
    ) -> dict[str, Any]:
        async with storage_manager.request(
            proxy_name,
            "POST",
            "folder/file/list",
            json={
                "volume": volume_name,
                "vfid": str(vfid),
                "relpath": relpath,
            },
        ) as (client_api_url, storage_resp):
            return await storage_resp.json()

    @staticmethod
    async def validate_model_definition(
        storage_manager: StorageSessionManager,
        model_vfolder_row: VFolderRow | Mapping[str, Any],
        model_definition_path: str | None,
    ) -> str | None:
        """
        Checks if model definition YAML exists and is syntactically perfect.
        Returns relative path to customized model-definition.yaml (if any) or None.
        """
        match model_vfolder_row:
            case VFolderRow():
                folder_name = model_vfolder_row.name
                vfid = model_vfolder_row.vfid
                folder_host = model_vfolder_row.host
            case _:
                folder_name = model_vfolder_row["name"]
                vfid = VFolderID(model_vfolder_row["quota_scope_id"], model_vfolder_row["id"])
                folder_host = model_vfolder_row["host"]

        proxy_name, volume_name = storage_manager.get_proxy_and_volume(folder_host)

        if model_definition_path:
            path = Path(model_definition_path)
            storage_reply = await ModelServicePredicateChecker._listdir(
                storage_manager, proxy_name, volume_name, vfid, path.parent.as_posix()
            )
            for item in storage_reply["items"]:
                if item["name"] == path.name:
                    yaml_name = model_definition_path
                    break
            else:
                raise InvalidAPIParameters(
                    f"Model definition YAML file {model_definition_path} not found inside the model storage"
                )
        else:
            storage_reply = await ModelServicePredicateChecker._listdir(
                storage_manager, proxy_name, volume_name, vfid, "."
            )
            model_definition_candidates = ["model-definition.yaml", "model-definition.yml"]
            for item in storage_reply["items"]:
                if item["name"] in model_definition_candidates:
                    yaml_name = item["name"]
                    break
            else:
                raise InvalidAPIParameters(
                    'Model definition YAML file "model-definition.yaml" or "model-definition.yml" not found inside the model storage'
                )

        chunks = bytes()
        async with storage_manager.request(
            proxy_name,
            "POST",
            "folder/file/fetch",
            json={
                "volume": volume_name,
                "vfid": str(vfid),
                "relpath": f"./{yaml_name}",
            },
        ) as (client_api_url, storage_resp):
            while True:
                chunk = await storage_resp.content.read(DEFAULT_CHUNK_SIZE)
                if not chunk:
                    break
                chunks += chunk
        model_definition_yaml = chunks.decode("utf-8")
        yaml = YAML()
        model_definition_dict = yaml.load(model_definition_yaml)
        try:
            model_definition = model_definition_iv.check(model_definition_dict)
            assert model_definition is not None
        except t.DataError as e:
            raise InvalidAPIParameters(
                f"Failed to validate model definition from vFolder {folder_name} (ID"
                f" {vfid.folder_id}): {e}",
            ) from e
        except YAMLError as e:
            raise InvalidAPIParameters(f"Invalid YAML syntax: {e}") from e

        return yaml_name


class EndpointStatistics:
    @classmethod
    async def batch_load_by_endpoint_impl(
        cls,
        redis_stat: RedisConnectionInfo,
        endpoint_ids: Sequence[UUID],
    ) -> Sequence[Optional[Mapping[str, Any]]]:
        async def _build_pipeline(redis: Redis) -> Pipeline:
            pipe = redis.pipeline()
            for endpoint_id in endpoint_ids:
                pipe.get(f"inference.{endpoint_id}.app")
            return pipe

        stats = []
        results = await redis_helper.execute(redis_stat, _build_pipeline)
        for result in results:
            if result is not None:
                stats.append(msgpack.unpackb(result))
            else:
                stats.append(None)
        return stats

    @classmethod
    async def batch_load_by_endpoint(
        cls,
        ctx: "GraphQueryContext",
        endpoint_ids: Sequence[UUID],
    ) -> Sequence[Optional[Mapping[str, Any]]]:
        return await cls.batch_load_by_endpoint_impl(ctx.redis_stat, endpoint_ids)

    @classmethod
    async def batch_load_by_replica(
        cls,
        ctx: GraphQueryContext,
        endpoint_replica_ids: Sequence[tuple[UUID, UUID]],
    ) -> Sequence[Optional[Mapping[str, Any]]]:
        async def _build_pipeline(redis: Redis) -> Pipeline:
            pipe = redis.pipeline()
            for endpoint_id, replica_id in endpoint_replica_ids:
                pipe.get(f"inference.{endpoint_id}.replica.{replica_id}")
            return pipe

        stats = []
        results = await redis_helper.execute(ctx.redis_stat, _build_pipeline)
        for result in results:
            if result is not None:
                stats.append(msgpack.unpackb(result))
            else:
                stats.append(None)
        return stats



from dataclasses import dataclass
from typing import Any, Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.user.actions.base import UserAction


@dataclass
class UserMonthStatsAction(UserAction):
    user_id: str

    @override
    def entity_id(self) -> Optional[str]:
        return self.user_id

    @override
    def operation_type(self):
        return "user_month_stats"


@dataclass
class UserMonthStatsActionResult(BaseActionResult):
    stats: list[Any]

    @override
    def entity_id(self) -> Optional[str]:
        return None



import logging

import sqlalchemy as sa

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.api.exceptions import ObjectNotFound
from ai.backend.manager.models.resource_policy import (
    KeyPairResourcePolicyRow,
)
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.services.keypair_resource_policy.actions.create_keypair_resource_policy import (
    CreateKeyPairResourcePolicyAction,
    CreateKeyPairResourcePolicyActionResult,
)
from ai.backend.manager.services.keypair_resource_policy.actions.delete_keypair_resource_policy import (
    DeleteKeyPairResourcePolicyAction,
    DeleteKeyPairResourcePolicyActionResult,
)
from ai.backend.manager.services.keypair_resource_policy.actions.modify_keypair_resource_policy import (
    ModifyKeyPairResourcePolicyAction,
    ModifyKeyPairResourcePolicyActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class KeypairResourcePolicyService:
    _db: ExtendedAsyncSAEngine

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
    ) -> None:
        self._db = db

    async def create_keypair_resource_policy(
        self, action: CreateKeyPairResourcePolicyAction
    ) -> CreateKeyPairResourcePolicyActionResult:
        to_store = action.creator.fields_to_store()

        async with self._db.begin_session() as db_sess:
            db_row = KeyPairResourcePolicyRow(**to_store)
            db_sess.add(db_row)
            result = db_row.to_dataclass()

        return CreateKeyPairResourcePolicyActionResult(keypair_resource_policy=result)

    async def modify_keypair_resource_policy(
        self, action: ModifyKeyPairResourcePolicyAction
    ) -> ModifyKeyPairResourcePolicyActionResult:
        name = action.name
        props = action.modifier

        async with self._db.begin_session() as db_sess:
            query = sa.select(KeyPairResourcePolicyRow).where(KeyPairResourcePolicyRow.name == name)
            result = await db_sess.execute(query)
            row: KeyPairResourcePolicyRow = result.scalar_one_or_none()
            if row is None:
                raise ObjectNotFound(f"Keypair resource policy with name {name} not found.")
            to_update = props.fields_to_update()
            for key, value in to_update.items():
                setattr(row, key, value)
            result = row.to_dataclass()

        return ModifyKeyPairResourcePolicyActionResult(keypair_resource_policy=result)

    async def delete_keypair_resource_policy(
        self, action: DeleteKeyPairResourcePolicyAction
    ) -> DeleteKeyPairResourcePolicyActionResult:
        name = action.name
        async with self._db.begin_session() as db_sess:
            query = sa.select(KeyPairResourcePolicyRow).where(KeyPairResourcePolicyRow.name == name)
            db_row = (await db_sess.execute(query)).scalar_one_or_none()
            if not db_row:
                raise ObjectNotFound(f"Keypair resource policy with name {name} not found.")
            await db_sess.delete(db_row)
            result = db_row.to_dataclass()

        return DeleteKeyPairResourcePolicyActionResult(keypair_resource_policy=result)



from dataclasses import dataclass
from typing import Any, Iterable, Optional, override

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.services.session.base import SessionAction


# TODO: Change this to BatchAction since it can destroy multiple sessions with recursive option
@dataclass
class DestroySessionAction(SessionAction):
    user_role: UserRole
    session_name: str
    forced: bool
    recursive: bool
    owner_access_key: AccessKey

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self):
        if self.recursive:
            return "destory_multi"
        return "destory"


@dataclass
class DestroySessionActionResult(BaseActionResult):
    destroyed_sessions: Iterable[SessionRow | BaseException]

    # TODO: Add proper type
    result: Any

    # TODO: Change this to `entity_ids`
    @override
    def entity_id(self) -> Optional[str]:
        return None



import uuid
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Optional, Self, override

from ai.backend.common.types import ResourceSlot, VFolderHostPermissionMap
from ai.backend.manager.models.group import GroupRow, ProjectType
from ai.backend.manager.types import Creator, OptionalState, PartialModifier, TriState


@dataclass
class GroupCreator(Creator):
    name: str
    domain_name: str
    type: Optional[ProjectType] = None
    description: Optional[str] = None
    is_active: Optional[bool] = None
    total_resource_slots: Optional[ResourceSlot] = None
    allowed_vfolder_hosts: Optional[dict[str, str]] = None
    integration_id: Optional[str] = None
    resource_policy: Optional[str] = None
    container_registry: Optional[dict[str, str]] = None

    def fields_to_store(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "domain_name": self.domain_name,
            "type": self.type,
            "description": self.description,
            "is_active": self.is_active,
            "total_resource_slots": self.total_resource_slots,
            "allowed_vfolder_hosts": self.allowed_vfolder_hosts,
            "integration_id": self.integration_id,
            "resource_policy": self.resource_policy,
            "container_registry": self.container_registry,
        }


@dataclass
class GroupData:
    id: uuid.UUID = field(compare=False)
    name: str
    description: Optional[str]
    is_active: bool
    created_at: datetime = field(compare=False)
    modified_at: datetime = field(compare=False)
    integration_id: Optional[str]
    domain_name: str
    total_resource_slots: ResourceSlot
    allowed_vfolder_hosts: VFolderHostPermissionMap
    dotfiles: bytes
    resource_policy: str
    type: ProjectType
    container_registry: Optional[dict[str, str]]

    @classmethod
    def from_row(cls, row: Optional[GroupRow]) -> Optional[Self]:
        if row is None:
            return None
        return cls(
            id=row.id,
            name=row.name,
            description=row.description,
            is_active=row.is_active,
            created_at=row.created_at,
            modified_at=row.modified_at,
            integration_id=row.integration_id,
            domain_name=row.domain_name,
            total_resource_slots=row.total_resource_slots,
            allowed_vfolder_hosts=row.allowed_vfolder_hosts,
            dotfiles=row.dotfiles,
            resource_policy=row.resource_policy,
            type=row.type,
            container_registry=row.container_registry,
        )


@dataclass
class GroupModifier(PartialModifier):
    name: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    description: TriState[str] = field(
        default_factory=TriState[str].nop,
    )
    is_active: OptionalState[bool] = field(default_factory=OptionalState[bool].nop)
    domain_name: OptionalState[str] = field(
        default_factory=OptionalState[str].nop,
    )
    total_resource_slots: OptionalState[ResourceSlot] = field(
        default_factory=OptionalState[ResourceSlot].nop
    )
    allowed_vfolder_hosts: OptionalState[dict[str, str]] = field(
        default_factory=OptionalState[dict[str, str]].nop
    )
    integration_id: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    resource_policy: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    container_registry: TriState[dict[str, str]] = field(
        default_factory=TriState[dict[str, str]].nop
    )

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.description.update_dict(to_update, "description")
        self.is_active.update_dict(to_update, "is_active")
        self.domain_name.update_dict(to_update, "domain_name")
        self.total_resource_slots.update_dict(to_update, "total_resource_slots")
        self.allowed_vfolder_hosts.update_dict(to_update, "allowed_vfolder_hosts")
        self.integration_id.update_dict(to_update, "integration_id")
        self.resource_policy.update_dict(to_update, "resource_policy")
        self.container_registry.update_dict(to_update, "container_registry")
        return to_update



from __future__ import annotations

import logging
from pprint import pformat, pprint
from typing import Optional

import click
import sqlalchemy as sa
from redis.asyncio.client import Pipeline, Redis
from tabulate import tabulate

from ai.backend.common import redis_helper
from ai.backend.common.arch import CURRENT_ARCH
from ai.backend.common.docker import validate_image_labels
from ai.backend.common.exception import UnknownImageReference
from ai.backend.common.types import ImageAlias
from ai.backend.logging import BraceStyleAdapter

from ..data.image.types import ImageStatus
from ..models.image import ImageAliasRow, ImageIdentifier, ImageRow
from ..models.image import rescan_images as rescan_images_func
from ..models.utils import connect_database
from .context import CLIContext, redis_ctx

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


async def list_images(cli_ctx, short, installed_only):
    # Connect to postgreSQL DB
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_readonly_session() as session,
        redis_ctx(cli_ctx) as redis_conn_set,
    ):
        displayed_items = []
        try:
            # Idea: Add `--include-deleted` option to include deleted images?
            items = await ImageRow.list(session)
            # NOTE: installed/installed_agents fields are no longer provided in CLI,
            #       until we finish the epic refactoring of image metadata db.
            if installed_only:

                async def _build_scard_pipeline(redis: Redis) -> Pipeline:
                    pipe = redis.pipeline()
                    for item in items:
                        await pipe.scard(item.name)
                    return pipe

                installed_counts = await redis_helper.execute(
                    redis_conn_set.image, _build_scard_pipeline
                )
                installed_items = []

                async def _build_smembers_pipeline(redis: Redis) -> Pipeline:
                    pipe = redis.pipeline()
                    for item, installed_count in zip(items, installed_counts):
                        if installed_count > 0:
                            installed_items.append(item)
                            await pipe.smembers(item.name)
                    return pipe

                agents_per_installed_items = await redis_helper.execute(
                    redis_conn_set.image,
                    _build_smembers_pipeline,
                )
                for item, installed_agents in zip(installed_items, agents_per_installed_items):
                    formatted_installed_agents = " ".join(
                        map(lambda s: s.decode(), installed_agents)
                    )
                    if short:
                        displayed_items.append((
                            item.image_ref.canonical,
                            item.config_digest,
                            formatted_installed_agents,
                        ))
                    else:
                        print(f"{pformat(item)} @ {formatted_installed_agents}")
            else:
                for item in items:
                    if short:
                        displayed_items.append((item.image_ref.canonical, item.config_digest))
                    else:
                        pprint(item)
            if short:
                print(tabulate(displayed_items, tablefmt="plain"))
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def inspect_image(cli_ctx, canonical_or_alias, architecture):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_readonly_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(canonical_or_alias, architecture),
                    ImageAlias(canonical_or_alias),
                ],
            )
            pprint(await image_row.inspect())
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def forget_image(cli_ctx, canonical_or_alias, architecture):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(canonical_or_alias, architecture),
                    ImageAlias(canonical_or_alias),
                ],
            )
            await image_row.mark_as_deleted(session)
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def purge_image(cli_ctx, canonical_or_alias, architecture):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(canonical_or_alias, architecture),
                    ImageAlias(canonical_or_alias),
                ],
                filter_by_statuses=None,
            )
            await session.delete(image_row)
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def set_image_resource_limit(
    cli_ctx,
    canonical_or_alias,
    slot_type,
    range_value,
    architecture,
):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(canonical_or_alias, architecture),
                    ImageAlias(canonical_or_alias),
                ],
            )
            await image_row.set_resource_limit(slot_type, range_value)
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def rescan_images(
    cli_ctx: CLIContext, registry_or_image: str, project: Optional[str] = None
) -> None:
    if not registry_or_image:
        raise click.BadArgumentUsage("Please specify a valid registry or full image name.")
    async with (
        connect_database(cli_ctx.local_config) as db,
    ):
        try:
            result = await rescan_images_func(db, registry_or_image, project)
            for error in result.errors:
                log.error(f"Failed to scan registries: {error}")
        except Exception as e:
            log.exception(f"Unknown error occurred. Error: {e}")


async def alias(cli_ctx, alias, target, architecture):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(target, architecture),
                ],
            )
            await ImageAliasRow.create(session, alias, image_row)
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


async def dealias(cli_ctx, alias):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_session() as session,
    ):
        alias_row = await session.scalar(
            sa.select(ImageAliasRow).where(ImageAliasRow.alias == alias),
        )
        if alias_row is None:
            log.exception("Alias not found.")
            return
        await session.delete(alias_row)


async def validate_image_alias(cli_ctx, alias: str) -> None:
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_readonly_session() as session,
    ):
        try:
            image_row = await ImageRow.from_alias(session, alias)
            for key, value in validate_image_labels(image_row.labels).items():
                print(f"{key:<40}: ", end="")
                if isinstance(value, list):
                    value = f"[{', '.join(value)}]"
                print(value)

        except UnknownImageReference:
            log.error(f"No images were found with alias: {alias}")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")


def _resolve_architecture(current: bool, architecture: Optional[str]) -> str:
    if architecture is not None:
        return architecture
    if current:
        return CURRENT_ARCH

    raise ValueError("Unreachable code!")


async def validate_image_canonical(
    cli_ctx, canonical: str, current: bool, architecture: Optional[str] = None
) -> None:
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_readonly_session() as session,
    ):
        try:
            if current or architecture is not None:
                resolved_arch = _resolve_architecture(current, architecture)
                image_row = await ImageRow.resolve(
                    session, [ImageIdentifier(canonical, resolved_arch)]
                )

                print(f"{'architecture':<40}: {resolved_arch}")
                for key, value in validate_image_labels(image_row.labels).items():
                    print(f"{key:<40}: ", end="")
                    if isinstance(value, list):
                        value = f"{', '.join(value)}"
                    print(value)
            else:
                rows = await session.scalars(
                    sa.select(ImageRow).where(
                        sa.and_(ImageRow.name == canonical, ImageRow.status == ImageStatus.ALIVE)
                    )
                )
                image_rows = rows.fetchall()
                if not image_rows:
                    raise UnknownImageReference(f"{canonical}")
                for i, image_row in enumerate(image_rows):
                    if i > 0:
                        print("-" * 50)
                    print(f"{'architecture':<40}: {image_row.architecture}")
                    for key, value in validate_image_labels(image_row.labels).items():
                        print(f"{key:<40}: ", end="")
                        if isinstance(value, list):
                            value = f"{', '.join(value)}"
                        print(value)

        except UnknownImageReference as e:
            log.error(f"{e}")
        except Exception as e:
            log.exception(f"An error occurred. Error: {e}")



from dataclasses import dataclass, field
from typing import Any, Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.models.user import UserRole, UserStatus
from ai.backend.manager.services.user.actions.base import UserAction
from ai.backend.manager.services.user.type import UserData
from ai.backend.manager.types import OptionalState, PartialModifier, TriState


@dataclass
class UserModifier(PartialModifier):
    username: OptionalState[str] = field(default_factory=OptionalState.nop)
    password: OptionalState[str] = field(default_factory=OptionalState.nop)
    need_password_change: OptionalState[bool] = field(default_factory=OptionalState.nop)
    full_name: OptionalState[str] = field(default_factory=OptionalState.nop)
    description: OptionalState[str] = field(default_factory=OptionalState.nop)
    is_active: OptionalState[bool] = field(default_factory=OptionalState.nop)
    status: OptionalState[UserStatus] = field(default_factory=OptionalState.nop)
    domain_name: OptionalState[str] = field(default_factory=OptionalState.nop)
    role: OptionalState[UserRole] = field(default_factory=OptionalState.nop)
    allowed_client_ip: TriState[list[str]] = field(default_factory=TriState.nop)
    totp_activated: OptionalState[bool] = field(default_factory=OptionalState.nop)
    resource_policy: OptionalState[str] = field(default_factory=OptionalState.nop)
    sudo_session_enabled: OptionalState[bool] = field(default_factory=OptionalState.nop)
    main_access_key: TriState[str] = field(default_factory=TriState.nop)
    container_uid: TriState[int] = field(default_factory=TriState.nop)
    container_main_gid: TriState[int] = field(default_factory=TriState.nop)
    container_gids: TriState[list[int]] = field(default_factory=TriState.nop)

    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.username.update_dict(to_update, "username")
        self.password.update_dict(to_update, "password")
        self.need_password_change.update_dict(to_update, "need_password_change")
        self.full_name.update_dict(to_update, "full_name")
        self.description.update_dict(to_update, "description")
        self.is_active.update_dict(to_update, "is_active")
        self.status.update_dict(to_update, "status")
        self.domain_name.update_dict(to_update, "domain_name")
        self.role.update_dict(to_update, "role")
        self.allowed_client_ip.update_dict(to_update, "allowed_client_ip")
        self.totp_activated.update_dict(to_update, "totp_activated")
        self.resource_policy.update_dict(to_update, "resource_policy")
        self.sudo_session_enabled.update_dict(to_update, "sudo_session_enabled")
        self.main_access_key.update_dict(to_update, "main_access_key")
        self.container_uid.update_dict(to_update, "container_uid")
        self.container_main_gid.update_dict(to_update, "container_main_gid")
        self.container_gids.update_dict(to_update, "container_gids")
        return to_update


@dataclass
class ModifyUserAction(UserAction):
    email: str
    modifier: UserModifier
    group_ids: OptionalState[list[str]] = field(default_factory=OptionalState.nop)

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self) -> str:
        return "modify"


@dataclass
class ModifyUserActionResult(BaseActionResult):
    data: Optional[UserData]
    success: bool

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.data.id) if self.data else None



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseAction, BaseBatchAction


@dataclass
class SessionAction(BaseAction):
    @override
    def entity_type(self):
        return "session"


@dataclass
class SessionBatchAction(BaseBatchAction):
    @override
    def entity_type(self):
        return "session"



import logging
import urllib.parse
from typing import AsyncIterator, override

import aiohttp

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.exceptions import ContainerRegistryProjectEmpty

from .base import (
    BaseContainerRegistry,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class GitLabRegistry(BaseContainerRegistry):
    @override
    async def fetch_repositories(self, sess: aiohttp.ClientSession) -> AsyncIterator[str]:
        if not self.registry_info.project:
            raise ContainerRegistryProjectEmpty(self.registry_info.type, self.registry_info.project)

        access_token = self.registry_info.password
        api_endpoint = self.registry_info.extra.get("api_endpoint", None)

        if api_endpoint is None:
            raise RuntimeError('"api_endpoint" is not provided for GitLab registry!')

        encoded_project_id = urllib.parse.quote(self.registry_info.project, safe="")
        repo_list_url = f"{api_endpoint}/api/v4/projects/{encoded_project_id}/registry/repositories"

        headers = {
            "Accept": "application/json",
            "PRIVATE-TOKEN": access_token,
        }
        page = 1

        while True:
            async with sess.get(
                repo_list_url,
                headers=headers,
                params={"per_page": 30, "page": page},
            ) as response:
                if response.status == 200:
                    data = await response.json()

                    for repo in data:
                        yield repo["path"]
                    if "next" in response.headers.get("Link", ""):
                        page += 1
                    else:
                        break
                else:
                    raise RuntimeError(
                        f"Failed to fetch repositories for project {self.registry_info.project}! {response.status} error occurred."
                    )



from dataclasses import dataclass
from typing import Callable, Optional, Protocol, Self

import sqlalchemy as sa

from .exceptions import EmptySQLCondition

type QuerySingleCondition = sa.sql.expression.BinaryExpression

type QueryOptionCallable = Callable[[sa.sql.Select], sa.sql.Select]
type QueryOption = Callable[..., Callable[[sa.sql.Select], sa.sql.Select]]


@dataclass
class QueryCondition:
    """
    A class representing a SQL condition for querying.
    Raises an EmptySQLCondition exception if the condition is empty.
    """

    _sql_condition: Optional[QuerySingleCondition]

    @classmethod
    def multiple(cls, operator: Callable, conditions: list[Self]) -> Self:
        """
        Args:
            operator: The SQL operator to use for combining conditions, one of `sqlalchemy.or_`, `sqlalchemy.and_`.
            conditions: A list of QueryCondition instances to combine.
        Returns:
            A new QueryCondition instance with the combined conditions.
        """
        sql_conditions = []
        for cond in conditions:
            try:
                sql_conditions.append(cond.final_sql_condition)
            except EmptySQLCondition:
                continue
        if not sql_conditions:
            return cls(None)
        if len(sql_conditions) == 1:
            return cls(sql_conditions[0])
        return cls(operator(*sql_conditions))

    @classmethod
    def single(cls, condition: QuerySingleCondition) -> Self:
        return cls(condition)

    @property
    def final_sql_condition(self) -> QuerySingleCondition:
        if self._sql_condition is None:
            raise EmptySQLCondition
        return self._sql_condition


class _LoadableField(Protocol):
    def __call__(self, *args, **kwargs) -> Callable:
        pass


class _JoinableField(Protocol):
    def __call__(self, *args, **kwargs) -> sa.orm.attributes.InstrumentedAttribute:
        pass


def load_related_field(field: _LoadableField) -> QueryOptionCallable:
    return lambda stmt: stmt.options(field())


def join_by_related_field(field: _JoinableField) -> QueryOptionCallable:
    return lambda stmt: stmt.join(field())
